# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import abc
import os
from requests import Session
from arjuna.tpi.error import HttpUnexpectedStatusCodeError, HttpSendError, HttpConnectError
from requests.exceptions import ConnectionError, TooManyRedirects, ProxyError, InvalidProxyURL
import time

from .request import _HttpRequest
from .response import HttpResponse
from .cookie import HttpCookie
from arjuna.interact.http.model.message import HttpMessage

class HttpCookie:

    def __init__(self, session, cookie):
        self.__cookie = cookie

    @property
    def _cookie(self):
        return self.__cookie

    @property
    def value(self):
        return self._cookie.value

    @property
    def secure(self):
        return self._cookie.secure

    @property
    def httponly(self):
        return self._cookie.has_nonstandard_attr("HttpOnly")


class HttpSession:
    '''
        Create an HTTP Session. Does automatic cookie management.

        Keyword Arguments:
            url: Base URL for this HTTP session. If relative path is used as a route in sender methods like `.get`, then this URL is prefixed to their provided routes.
            oauth_token: OAuth 2.0 Bearer token for this session.
            request_content_handler: Default content type handler for requests sent in this session. Overridable in individual sender methods. Default is Http.content.urlencoded.
            headers: HTTP headers to be added to request headers made by this session.
            max_redirects: Maximum number of redirects allowed for a request. Default is 30.
            auth: HTTP Authentication object: Basic/Digest.
            proxy: Proxies dict to be associated with this session.
    '''

    def __init__(self, *, url=None, oauth_token=None, request_content_handler=None, headers=None, max_redirects=None, auth=None, proxy=None, _auto_session=True):
        self.__url = url is not None and url.strip() or None
        self.__request_content_handler = request_content_handler
        from .http import Http
        if self.__request_content_handler is None:
            self.__request_content_handler = Http.content.urlencoded
        self.__session = None
        self.__provided_headers = headers
        if _auto_session:
            self._set_session(Session())
            if max_redirects is not None:
                self.__session.max_redirects = max_redirects
            if auth is not None:
                self.__session.auth = auth
            if proxy is not None:
                self.__session.proxies = proxy
            else:
                from arjuna import C, Http
                if C("http.proxy.enabled"):
                    self.__session.proxies = Http.proxy(C('http.proxy.host'), C('http.proxy.port'))

        if oauth_token:
            self.__session.headers['Authorization'] = f'Bearer {oauth_token}'

    @property
    def cookies(self) -> dict:
        '''
            All current cookie name/values in this session object.
        '''
        return self.__session.cookies.get_dict()

    @property
    def parsed_cookies(self) -> dict:
        '''
            All current cookie name and corresponding `HttpCookie` object in this session object.
        '''
        return {cookie.name:HttpCookie(self, cookie) for cookie in self._session.cookies}

    def add_cookies(self, cookie_dict):
        '''
            Add cookies to the session.
        '''
        self.__session.cookies.update(cookie_dict)

    @property
    def request_content_handler(self):
        '''
        Request content handler for content formatting.
        '''
        return self.__request_content_handler

    def _set_session(self, session):
        self.__session = session
        if self.__provided_headers is not None:
            self.__session.headers.update(self.__provided_headers)
        from .http import Http
        if self.request_content_handler != Http.content.custom:
            self.__session.headers['Content-Type'] = Http.content.get_content_type(self.request_content_handler)

    @property
    def headers(self):
        '''
        Request headers.
        '''
        return self._session.headers

    @property
    def auth(self):
        '''
        HTTP Authentication object.
        '''
        return self._session.auth

    @property
    def url(self):
        '''
            Base URL for this session.
        '''
        return self.__url

    @property
    def _session(self):
        return self.__session

    @property
    def _request_headers(self):
        return self.__session.headers

    def __route(self, route):
        route = route.strip()
        if route.lower().startswith("http"):
            return route
        else:
            if route.startswith("/"):
                return self.url + route
            else:
                return self.url + "/" + route

    def __register_network_info(self, request, response):
        from arjuna import Arjuna
        from arjuna.tpi.helper.arjtype import NetworkPacketInfo
        # Should be configurable
        sub_network_packets = []
        for redir_resp in response.redir_history:
            redir_req = redir_resp.request
            sub_network_packets.append(
                NetworkPacketInfo(
                    label="Sub-Request: {} {}".format(redir_req.method, redir_req.url), 
                    request=str(redir_req), 
                    response=str(redir_resp),
                    sub_network_packets=tuple()
                )
            )

        # The request for last response object was the last request and hence the last redirection.
        if response.redir_history:
            # last_req = response.last_request
            # if not last_req:
            last_req = response.request
            sub_network_packets.append(
                                NetworkPacketInfo(
                                    label="Sub-Request: {} {}".format(last_req.method, last_req.url), 
                                    request=str(last_req), 
                                    response=str(response),
                                    sub_network_packets=tuple()
                                )
                            )                

        Arjuna.get_report_metadata().add_network_packet_info(
            NetworkPacketInfo(label=request.label, request=str(request), response=str(response), sub_network_packets=tuple(sub_network_packets))
        )

    def message(self, msg_name=None, **fargs) -> HttpResponse:
        if msg_name is not None:
            msg = HttpMessage.from_file(self, msg_name, **fargs)
        else:
            msg = HttpMessage.root(self)
        return msg.send()

    def _validate_status_code(cls, response, codes, *, expected=True):
        if expected:
            if response.status_code not in codes:
                raise HttpUnexpectedStatusCodeError(response.request, response, msg=f"Expected code(s): {codes}")
        else:
            if response.status_code in codes:
                raise HttpUnexpectedStatusCodeError(response.request, response, msg=f"Unexpected code(s): {codes}")


    def _send(self, request) -> HttpResponse:
        '''
            Send the provided HttpRequest to server.

            In case of ConnectionError, retries the connection 5 times at a gap of 1 second. Currently, not configurable.

            Returns
                `HttpResponse` object. In case of redirections, this is the last HttpResponse object, which encapsulates all redirections which can be retrieved from it.
        '''
        from arjuna import Arjuna, log_info
        from arjuna.tpi.helper.arjtype import NetworkPacketInfo
        log_info(request.label)
        max_connection_retries = 5
        try:
            counter = 0
            exc_flag = False
            exc_desc = None
            while counter < max_connection_retries:
                counter += 1
                try:
                    if self._session.proxies:
                        response = HttpResponse(self, self._session.send(request._request, allow_redirects=request.allow_redirects, timeout=request.timeout, proxies=self._session.proxies, verify=False))
                    else:
                        response = HttpResponse(self, self._session.send(request._request, allow_redirects=request.allow_redirects, timeout=request.timeout))
                except (ProxyError, InvalidProxyURL) as e:
                    raise HttpConnectError(request, "There is an error in connecting to the configured proxy. Proxy settings: {}. Error: {}".format(self.__session.proxies, str(e)))
                except ConnectionError as f:
                    exc_flag = True
                    exc_desc = str(f)
                    time.sleep(1)
                    continue
                else:
                    break
            if exc_flag:
                raise HttpConnectError(request, "Connection error despite trying 5 times. Error: {}".format(exc_desc))
        except TooManyRedirects as e:
            response = HttpResponse(self._session, e.response)
            self.__register_network_info(request, response)
            raise HttpSendError(self, response, str(e) + ". Error redir URL: " + e.response.url)
        except Exception as e:
            import traceback
            response = "Error in sending the request\n"
            response += e.__class__.__name__ + ":" + str(e) + "\n"
            response += traceback.format_exc()
            Arjuna.get_report_metadata().add_network_packet_info(
                NetworkPacketInfo(label=request.label, request=str(request), response=str(response), sub_network_packets=tuple())
            )
            raise e
        else:
            self.__register_network_info(request, response)
            if request.xcodes is not None:
                self._validate_status_code(response, request.xcodes)
            return response


    def get(self, route, label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP GET request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        request = _HttpRequest(self, self.__route(route), method="get", label=label, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)
        return self._send(request)


    def head(self, route, label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP HEAD request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        request = _HttpRequest(self, self.__route(route), method="head", label=label, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)
        return self._send(request)


    def delete(self, route, label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP DELETE request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        request = _HttpRequest(self, self.__route(route), method="delete", label=label, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)
        return self._send(request)

    def post(self, route, *, content, label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP POST request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            content: Content to be sent in this HTTP request. If passed as string, then content-type set in session is used using the content request handler. It can also be a dictionary with keys - 'content' and 'type'.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        request = _HttpRequest(self, self.__route(route), method="post", label=label, content=content, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)
        return self._send(request)

    def put(self, route, *, content, label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP PUT request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            content: Content to be sent in this HTTP request. If passed as string, then content-type set in session is used using the content request handler. It can also be a dictionary with keys - 'content' and 'type'.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        request = _HttpRequest(self, self.__route(route), method="put", label=label, content=content, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)
        return self._send(request)

    def patch(self, route, *, content, label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP PUT request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            content: Content to be sent in this HTTP request. If passed as string, then content-type set in session is used using the content request handler. It can also be a dictionary with keys - 'content' and 'type'.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        request = _HttpRequest(self, self.__route(route), method="patch", label=label, content=content, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)
        return self._send(request)

    def options(self, route, *, content, label=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP PUT request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            content: Content to be sent in this HTTP request. If passed as string, then content-type set in session is used using the content request handler. It can also be a dictionary with keys - 'content' and 'type'.
            xcodes: Expected HTTP response code(s).
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        request = _HttpRequest(self, self.__route(route), method="options", label=label, content=content, xcodes=xcodes, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)
        return self._send(request)






# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .generator import _gen

class _DataEntity:
    _MANDATORY = set()
    _DEFAULT = set()
    _REFDICT = dict()

def _init(self, **kwargs):
    wrong_args = False
    provided_set = kwargs.keys()
    mandatory_args_missing = self._MANDATORY - provided_set
    if mandatory_args_missing:
        wrong_args = True
    default_arg_diff = kwargs.keys() - self._REFDICT.keys()
    if default_arg_diff:
        wrong_args = True

    if wrong_args:
        mandatory_args_missing = mandatory_args_missing and f" Missing mandatory args: {mandatory_args_missing}. Must: {self._MANDATORY}" or ""
        allowed = set(self._REFDICT.keys())
        default_arg_diff = default_arg_diff and f" Unexpected args: {default_arg_diff}. Allowed: {allowed}." or ""
        raise Exception(f"Wrong arguments.{mandatory_args_missing}{default_arg_diff}")

    def update_gen_data(key, value):
        if hasattr(value, '__call__'):
            value = value()
        elif isinstance(value, _gen):
            value = value.generate()

        vars(self)[key] = value

    final_dict = dict()
    final_dict.update(self._REFDICT)
    final_dict.update(kwargs)    

    for k,v in final_dict.items():
        update_gen_data(k,v)

def _attr(self, name):
    if name in self._REFDICT:
        return self._REFDICT[name]
    else:
        raise AttributeError("{} object does not have an attribute with name: {}.".format(self.__class__.__name__, name))

def _as_dict(self, remove_none=True):
    d = dict(vars(self))
    if remove_none:
        return {k:v for k,v in d.items() if v is not None}
    else:
        return d

def _str(self):
    return "{klass}({attrs})".format(
        klass = self.__class__.__name__,
        attrs = ", ".join("{}={}".format(k,v) for k,v in self.as_dict().items())
    )

def _iter(self):
    return iter(self.as_dict())

def __getitem__(self, k):
    return getattr(self, k)

def data_entity(entity_name, *attrs, bases=tuple(), **attrs_with_defaults):
    '''
        Create a new Data Entity class with provided name and attributes.

        Arguments:
            entity_name: The class name for this new Data Entity type.
            *attrs: Arbitrary names for Python attributes to be associated with objects of this entity.

        Keyword Arguments:
            bases: Base data entities for this entity. Can be a string or tuple or list.
            **attrs_with_defaults: Arbitrary attributes to be associated with objects of this entity, with the defaults that are provided.

        Note:
            The defaults that are provided can be any of the following:
                * Any Python Object
                * A Python callable
                * Arjuna `generator`
                * Arjuna `composite`

        Note:
            When you provide one or more bases, the overriding order is B1 -> B2 -> B3 ..... -> This Entity.

            At each stage of this chain, you can

                - Add one or more mandatory attributes.
                - Add one or more optional attributes
                - Make an optional attribute in base entity as mandatory.
                - Make a mandatory attribute as optional by assigning a default value.
                - Change the default for an existing default attribute to something else.
    '''
    for attr in attrs:
        if type(attr) is not str:
            raise Exception(f"Data Entity fields must be valid Python name strings. Got: {attr}")

    final_attrs = []
    # In any multiple names in single string are provided like a namedtuple, added as individual names.
    for attr in attrs:
        final_attrs.extend(attr.split())

    if not set(final_attrs).isdisjoint(set(attrs_with_defaults.keys())):
        raise Exception("Mandatory attribute can not be assigned a default value. Evaluate usage for: {}".format(set(final_attrs).intersection(set(attrs_with_defaults.keys()))))

    namespace = dict()
    namespace['_MANDATORY'] = set()
    namespace['_DEFAULT'] = set()
    namespace['_REFDICT'] = dict()

    if type(bases) not in {list, tuple}:
        bases = (bases,)

    for base in bases:
        base_vars = vars(base)
        namespace['_MANDATORY'] = namespace['_MANDATORY'].union(base_vars['_MANDATORY'])
        # Remove these params from Defaults
        for attr in namespace['_DEFAULT'].intersection(namespace['_MANDATORY']):
            namespace['_DEFAULT'].remove(attr)

        namespace['_DEFAULT'] = namespace['_DEFAULT'].union(base_vars['_DEFAULT'])
        # Remove optional params from Mandatory
        for attr in namespace['_DEFAULT'].intersection(namespace['_MANDATORY']):
            namespace['_MANDATORY'].remove(attr)
        #namespace['_MANDATORY'].difference_update(namespace['_DEFAULT'])
        namespace['_REFDICT'].update(base_vars['_REFDICT'])

    namespace['_MANDATORY'] = namespace['_MANDATORY'].union(set(final_attrs))
    # Remove mandatory params from Defaults
    for attr in namespace['_DEFAULT'].intersection(namespace['_MANDATORY']):
        namespace['_DEFAULT'].remove(attr)
    namespace['_DEFAULT'] = namespace['_DEFAULT'].union(set(attrs_with_defaults.keys()))
    # Remove optional params from Mandatory
    for attr in namespace['_DEFAULT'].intersection(namespace['_MANDATORY']):
        namespace['_MANDATORY'].remove(attr)
    namespace['_REFDICT'].update({f: None for f in final_attrs})
    namespace['_REFDICT'].update(attrs_with_defaults)

    namespace['__init__'] = _init
    namespace['__getattr__'] = _attr
    namespace['as_dict'] = _as_dict
    namespace['__str__'] = _str
    namespace['__iter__'] = _iter
    return type(entity_name, (_DataEntity,), namespace)




# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

##########################################################
# Raised and consumed by internal implementation of Arjuna
##########################################################

from arjuna.interact.gui.auto.finder.wmd import GuiWidgetMetaData

class ArjunaException(Exception):
    def __init__(self, message, scrreenshots_path=None, child=None):
        # child is child exception
        super().__init__(message)

        self.screenshot_path = scrreenshots_path
        self.child_Exceptions = []
        if child:
            try:
                raise child
            except ArjunaException as e:
                self.__extract_child_Exceptions(e);
                # } catch (java.lang.reflect.Invocation_target_exception f) {
                #     if (ArjunaException.class.is_assignable_from(f.get_target_exception().get_class())) {
                #         ArjunaException z = (ArjunaException) f.get_target_exception();
                #         extract_child_Exceptions(z);
                #     } else {
                #         insert_child_Exception(f);
                #     }
            except Exception as f:
                self.__insert_child_Exception(f);

    def contains_screenshot(self):
        return self.screenshot_path is not None

    def get_screenshot_path(self):
        return self.screenshot_path

    def set_screenshot_path(self, path):
        self.screenshot_path = path

    def __insert_child_Exception(self, e):
        self.child_Exceptions.append(e)

    def __extract_child_Exceptions(self, e):
        child_Exceptions = e.get_child_Exceptions()
        self.child_Exceptions.extend(child_Exceptions)

    def get_child_Exceptions(self):
        return self.child_Exceptions


def message_formatter(text, component=None, object_name=None, method_name=None, code=None):
    m = ""
    m = m + component and "{}::".format(component) or ""
    m = m + object_name and "{}::".format(object_name) or ""
    m = m + method_name and "{}::".format(method_name) or ""
    m = m + code and "{}::".format(code) or ""
    m = m + text
    return m


class Problem(ArjunaException):
    def __init__(self, text, screenshot_path, exc, component=None, object_name=None, method_name=None, code=None):
        super().__init__(message_formatter(text, component, object_name, method_name, code), exc, screenshot_path);
        self.problem_component = component;
        self.problem_object = object_name;
        self.problem_method = method_name;
        self.problem_code = code;

    def get_problem_component(self):
        return self.problem_component

    def get_problem_object(self):
        return self.problem_object

    def get_problem_method(self):
        return self.problem_method

    def get_problem_code(self):
        return self.problem_code

    def get_problem_text(self):
        return self.get_message()


class DirReaderFinishedException(Exception):
    def __init__(self):
        super().__init__("No more files.")


class UnsupportedRepresentationException(Exception):
    def __init__(self, klass_user_friendly_name, method, str_source_value, target_value_type):
        super().__init__(
            "Value.{}(): Can not represent {} types containing >>{}<< as {}.".format(method, str_source_value,
                                                                                     klass_user_friendly_name,
                                                                                     target_value_type))


class IncompatibleInputForValueException(Exception):
    def __init__(self, value, actual, value_type):
        super().__init__(
            "Incompatible input types >>{}<< (type: {}) supplied for creating {}.".format(value, actual, value_type))


class StringKeyValueContainerLookupException(Exception):
    def __init__(self,key):
        super().__init__("Invalid Key [{}] used for string key types container lookup.".format(key))

class WaitableError(Exception):

    def __init__(self, message):
        super().__init__(message)

def format_msg(msg):
    return msg and  "Error message: {}".format(msg) or ""

class _WidgetNotFoundError(WaitableError):

    def __init__(self, elem_name, *locators, container=None, relations=None, filters=None, message=None):
        container = container and  " in {}".format(container) or ""
        relations = relations and  " with relations {}".format(relations) or ""
        filters = filters and  " and filters {}".format(filters) or ""
        message = format_msg(message)
        super().__init__("{} not found using any of the locators: {}{}{}{}.{}".format(elem_name, GuiWidgetMetaData.locators_as_str(locators), relations, filters, container, message))

class _WidgetPresentError(WaitableError):

    def __init__(self, elem_type, *locators, message=None):
        message = message = format_msg(message)
        super().__init__("{} expected to be absent but still present for one of the locators: {}.{}".format(elem_type, GuiWidgetMetaData.locators_as_str(locators), message))

class GuiWidgetNotFoundError(_WidgetNotFoundError):

    def __init__(self, *locators, container=None, relations=None, filters=None, message=None):
        super().__init__("GuiWidget", *locators, container=container, relations=relations, filters=filters, message=message)

class _GuiWidgetPresentError(_WidgetPresentError):

    def __init__(self, *locators, message=None):
        super().__init__("GuiWidget", *locators, message=message)

class GuiWidgetNotReadyError(WaitableError):

    def __init__(self, message):
        super().__init__("GuiWidget is NOT ready for interaction. Tool message: {}".format(message))

class GuiWidgetTextNotSetError(WaitableError):

    def __init__(self, message):
        super().__init__(". Tool message: {}".format(message))

class ChildWindowNotFoundError(_WidgetNotFoundError):

    def __init__(self, *locators):
        super().__init__("Child window", *locators)

class ChildFrameNotFoundError(_WidgetNotFoundError):

    def __init__(self, *locators):
        super().__init__("Frame", *locators)

class ArjunaTimeoutError(WaitableError):

    def __init__(self, context, message):
        super().__init__(". Timeout in {}. Error Message: {}".format(context, message))  

class DataSourceFinished(StopIteration):
    def __init__(self, msg=None):
        super().__init__(msg is None and "Done" or msg)


class YamlError(Exception):
    '''
        Raised when there is an eror in Yaml structure or there is an error in its expected format in the context where it is used in Arjuna.
    '''

    def __init__(self, msg):
        super().__init__(msg)

class YamlUndefinedSectionError(Exception):
    '''
        Raised when the YamlList does not have the provided section key.
    '''

    def __init__(self, msg):
        super().__init__(msg)

class YamlListIndexError(Exception):
    '''
        Raised when the YamlList does not have the provided index.
    '''

    def __init__(self, msg):
        super().__init__(msg)

class TestGroupsFinished(Exception):
    pass

class ExclusionRuleMet(Exception):
    
    def __init__(self, rule):
        super().__init__("An exclusion rule was met for the object.")
        self.__rule = rule

    @property
    def rule(self):
        return self.__rule

class NoInclusionRuleMet(Exception):
    
    def __init__(self):
        super().__init__("None of the include rules were met.")


class InvalidSelectionRule(Exception):
    pass


class RulePatternDoesNotMatchError(Exception):

    def __init__(self, rule_str, pattern_class, expected_format):
        super().__init__(f"{rule_str} is not a {pattern_class.__name__}. Expected format: {expected_format}")



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from arjuna.tpi.engine.asserter import AsserterMixIn


class HttpPacket(AsserterMixIn, metaclass=abc.ABCMeta):

    def __init__(self, http_msg):
        super().__init__()
        self.__http_msg = http_msg

    @property
    def headers(self) -> dict:
        ''' 
            HTTP Headers for this message.
        '''
        return self.__http_msg.headers

    def assert_empty_content(self, *, msg):
        '''
            Validates if content is empty.

            Keyword Arguments:
                msg: A context string explaining why this assertion was done.
        '''
        if self.headers['Content-Length'] != '0':
            raise AssertionError("Content is not empty.")

    def assert_non_empty_content(self, *, msg):
        '''
            Validates if content is empty.

            Keyword Arguments:
                msg: A context string explaining why this assertion was done.
        '''
        if self.headers['Content-Length'] == '0':
            raise AssertionError("Content is empty.")

    def assert_header_match(self, header, value=None, *, msg):
        '''
            Validates the presence and optionally the value of an HTTP Header.

            Arguments:
                header: Name of header
                value: Text contained in header. If not provided, then only presence of header is checked.

            Keyword Arguments:
                msg: A context string explaining why this assertion was done.
        '''
        if header not in self.headers:
            raise AssertionError(f">>{header}<< not present in HTTP Request. {msg}")

        if value is not None:
            actual = self.headers[header]
            if actual != value:
                raise AssertionError(f"Value >>header<< is {actual} but was expected to be {value}. {msg}.")

    def assert_header_mismatch(self, header, value=None, *, msg):
        '''
            Validates the absence of header or mismatch of its content with provided value.

            Arguments:
                header: Name of header
                value: Text contained in header. If not provided, then only absence of header is checked.

            Keyword Arguments:
                msg: A context string explaining why this assertion was done.
        '''

        if value is None:
            if header in self.headers:
                raise AssertionError(f">>{header}<< is present in HTTP Request. {msg}")
        else:
            actual = self.headers[header]
            if actual == value:
                raise AssertionError(f"Value >>header<< is {actual}. It was expected to be different. {msg}.")




# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, auto

from arjuna.core.constant import GuiInteractionConfigType
from arjuna.tpi.helper.arjtype import Dictable
from arjuna.tpi.tracker import track


class _GuiInteractionConfigBuilder:

    def __init__(self):
        self.__settings = dict()

    def check_type(self, flag):
        self.__settings[GuiInteractionConfigType.CHECK_TYPE] = flag
        return self

    def check_pre_state(self, flag):
        self.__settings[GuiInteractionConfigType.CHECK_PRE_STATE] = flag
        return self

    def check_post_state(self, flag):
        self.__settings[GuiInteractionConfigType.CHECK_POST_STATE] = flag
        return self

    def scroll_to_view(self, flag):
        self.__settings[GuiInteractionConfigType.SCROLL_TO_VIEW] = flag
        return self

    def build(self):
        return GuiInteractionConfig(self.__settings)


class _GuiInteractionConfig:
    '''
        Stores configured values for behavior of actions on a given GuiElement.
    '''

    def __init__(self, settings):
        '''
            settings is a dict of dict of GuiInteractionConfigType key/value pairs
        '''
        # 
        self.__settings = settings

    @property
    def settings(self):
        '''
            Returns the dictionary of configured values for actions configuration.
        '''
        return self.__settings

    @staticmethod
    def builder():
        '''
            Returns a builder object to construct object of this class incrementally.
        '''
        return _GuiInteractionConfigBuilder()


@track("trace")
class GuiDriverExtendedConfig:

    def __init__(self, capabilities, browser_args, browser_prefs, browser_exts):
        self.__capabilities = capabilities
        self.__browser_args = browser_args
        self.__browser_prefs = browser_prefs
        self.__browser_exts = browser_exts

    @property
    def config(self):
        '''
            Returns all configuration settings as a single dictionary.
        '''
        map = dict()
        map["driverCapabilities"] = self.__capabilities
        map["browserArgs"] = self.__browser_args 
        map["browserPreferences"] = self.__browser_prefs
        map["browserExtensions"] = self.__browser_exts
        return map

@track("trace")
class GuiDriverExtendedConfigBuilder:

    def __init__(self):
        self.__capabilities = dict() 
        self.__browser_args = []
        self.__browser_prefs = dict()
        self.__browser_exts = []

    def capability(self, name, value):
        self.__capabilities[name] = value
        return self

    def browser_arg(self, arg):
        self.__browser_args.append(arg)
        return self

    def browser_pref(self, name, value):
        self.__browser_prefs[name] = value
        return self

    def browser_ext(self, path):
        self.__browser_exts.append(path)
        return self

    def build(self):
        return GuiDriverExtendedConfig(self.__capabilities, self.__browser_args, self.__browser_prefs, self.__browser_exts)


class Keys:

    class KeyChord:

        def __init__(self):
            self.__parts = []

        def text(self, text):
            self.__parts.append(text)
            return self

        def __getattr__(self, name):
            from selenium.webdriver.common.keys import Keys
            try:
                key_str = getattr(Keys, name.upper())
            except:
                raise AttributeError("There is no key with name {} defined.".format(name))
            else:
                def _dyn_key():
                    self.__parts.append(key_str)
                    return self
                return _dyn_key

        def key(self, key):
            self.__parts.append(key.name)
            return self

        @property
        def _parts(self):
            return self.__parts

    @classmethod
    def chord(cls):
        '''
            Returns a new KeyChord object.
        '''
        return cls.KeyChord()



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from .json.validator import ExpectedJsonValidator, UnexpectedJsonValidator
from .cookie.validator import CookieValidator
from .text.validator import ExpectedTextValidator, UnexpectedTextValidator
from .header.store import HeaderExtractor
from .text.store import TextExtractor
from .xml.store import XmlExtractor
from .validator import Validator
from arjuna.interact.http.model.internal.helper.yaml import convert_yaml_obj_to_content
from arjuna.tpi.engine.asserter import AsserterMixIn

class _HttpResProcessor(AsserterMixIn, metaclass=abc.ABCMeta):
    
    def __init__(self, session, codes=None, url=None, headers={}, cookies={}, has={}, match={}, store={}, validate={}):
        super().__init__()
        self.__session = session
        self.__match_set = False
        self.__has_set = False
        self.__store_set = False
        self.__validate_set = False
        if has:
            self.__has_set = True
        if match:
            self.__match_set = True
        if store:
            self.__store_set = True
        if validate:
            self.__validate_set = True

        self.__repr = {
            "codes" : None,
            "url" : url,
            "headers": headers,
            "cookies": cookies,
            "has": {
                "jpath": None,
                "regex": None,
                "partial": None
            },
            "match": {
                "jpath": None,
                "regex": None,
            },
            "store": store,
            "validate": validate
        }

        from arjuna.tpi.httpauto.request import _HttpRequest
        if codes is not None:
            self.__repr["codes"] = _HttpRequest._process_codes(codes)

        for k,v in has.items():
            if k in self.__repr["has"]:
                self.__repr["has"][k] = has[k]   

        for k,v in match.items():
            if k in self.__repr["match"]:
                self.__repr["match"][k] = match[k]   

        store_var_dict = {"jpath":{}, "regex":{}, "xpath":{}, "header":{}, "cookie":{}}
        import copy
        for k,v in store.items():
            self.__repr["store"][k] = copy.deepcopy(store_var_dict)
            for k2, v2 in v.items():
                if k2 in self.__repr["store"][k]:
                    self.__repr["store"][k][k2] = v2   

        for k in self.__repr["validate"]:
            if k not in self.__repr["store"]:
                raise Exception(f"{k} in validate section is not defined in store section of message.") 

    @property
    def _session(self):
        return self.__session

    @property
    def codes(self):
        return self.__repr["codes"]

    @property
    def url(self):
        return self.__repr["url"]

    @property
    def headers(self):
        return self.__repr["headers"]

    @property
    def cookies(self):
        return self.__repr["cookies"]

    @property
    def match(self):
        if self.__match_set:
            return self.__repr["match"]

    @property
    def has(self):
        if self.__has_set:
            return self.__repr["has"]

    @property
    def store(self):
        if self.__store_set:
            return self.__repr["store"]

    @property
    def validations(self):
        if self.__validate_set:
            return self.__repr["validate"]

    @abc.abstractmethod
    def _validate(self, response):
        pass

    @abc.abstractmethod
    def _get_text_validator(self, response):
        pass

    @abc.abstractmethod
    def _get_json_validator(self, response):
        pass

    def validate(self, response):
        self._validate(response)
        if self.has:
            if self.has["jpath"]:
                self._get_json_validator(response).assert_has_patterns(self.has["jpath"])
            if self.has["regex"]:
                self._get_text_validator(response).assert_has_patterns(self.has["regex"])
        if self.match:
            if self.match["jpath"]:
                self._get_json_validator(response).assert_match_for_patterns(self.match["jpath"])
            if self.match["regex"]:
                self._get_text_validator(response).assert_match_for_patterns(self.match["regex"])
        if self.store:
            for name, sdict in self.store.items():
                if sdict["header"]:
                    HeaderExtractor(response).store(name, sdict["header"])
                elif sdict["regex"]:
                    TextExtractor(response).store(name, sdict["regex"])
                elif sdict["xpath"]:
                    XmlExtractor(response).store(name, sdict["xpath"])
        if self.validations:
            validator = Validator(response)
            for k, target in self.validations.items():
                validator.validate(k, target)

class HttpExpectedResProcessor(_HttpResProcessor):
    
    def __init__(self, session, conditions_dict):
        super().__init__(session, **conditions_dict)
        self.__text_validator = None
        self.__json_validator = None

    def _get_text_validator(self, response):
        if self.__text_validator is None:
            self.__text_validator = ExpectedTextValidator(response.text) 
        return self.__text_validator

    def _get_json_validator(self, response):
        if self.__json_validator is None:
            self.__json_validator = ExpectedJsonValidator(response.json) 
        return self.__json_validator

    def _validate(self, response):
        if self.codes:
            self._session._validate_status_code(response, self.codes)
        if self.url:
            assert response.request.url == self.url 
        if self.headers:
            headers = {k:str(v) for k,v in  convert_yaml_obj_to_content(self.headers).items()}
            response.assert_headers(headers, msg="One or more headers do not match expected values.") 
        if self.cookies:
            CookieValidator(self._session, convert_yaml_obj_to_content(self.cookies))

class HttpUnexpectedResProcessor(_HttpResProcessor):

    def __init__(self, session, conditions_dict):
        super().__init__(session, **conditions_dict)
        self.__text_validator = None
        self.__json_validator = None

    def _get_text_validator(self, response):
        if self.__text_validator is None:
            self.__text_validator = UnexpectedTextValidator(response.text) 
        return self.__text_validator

    def _get_json_validator(self, response):
        if self.__json_validator is None:
            self.__json_validator = UnexpectedJsonValidator(response.json) 
        return self.__json_validator

    def _validate(self, response):
        if self.codes:
            self._session._validate_status_code(response, self.codes, expected=False)
        if self.headers:
            headers = {k:str(v) for k,v in  convert_yaml_obj_to_content(self.headers).items()}
            for k,v in headers.items():
                if k in response.headers:
                    self.asserter.assert_not_equal(response.headers[k], v, f"HTTP response header >>{k}<< value matches >>{v}<< which is unexpected.")



from enum import Enum
from arjuna.tpi.helper.arjtype import CIStringDict

def repr_dict(d, replace_value_enum=False):
    def format_key(k):
        if isinstance(k, Enum):
            return k.name.lower()
        elif type(k) is str:
            return k.lower()
        else:
            return k

    def format_value(v):
        if isinstance(v, Enum):
            if replace_value_enum:
                return v.name.lower()
            else:
                return v
        elif isinstance(v, CIStringDict):
            return repr_dict(v)
        else:
            return v
    return "{" + ", ".join(["{}={}".format(format_key(k), format_value(v)) for k,v in d.items()]) + "}"


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import pprint
from enum import Enum
from arjuna.tpi.helper.arjtype import *
from functools import partial

from weakref import WeakKeyDictionary


class EnumSwitch:
    def __init__(self, cases, const_args):
        self.dispatch_dict = cases
        self.const_args = const_args

    def __call__(self, econst):
        if econst in self.dispatch_dict:
            f_tuple = self.dispatch_dict[econst]
            arg_list = list(self.const_args) + list(f_tuple[1:])
            return f_tuple[0](*arg_list)



class PropertyEnum(Enum):
    pass
    # def _generate_next_value_(name, start, count, last_values):
    #     return ".".join(name.lower().split("_"))


class Descriptor(metaclass=abc.ABCMeta):
    def __init__(self):
        self._name = None
        self._values = WeakKeyDictionary()

    def __get__(self, instance, cls):
        return self._values.get(instance)

    def __set__(self, instance, value):
        self.validate(instance, value)
        if instance in self._values:
            self._values[instance].set_value(value)
        else:
            self._values[instance] = self._create_value(value)

    def set_name(self, name):
        self._name = name

    @abc.abstractmethod
    def _create_value(self, value):
        pass

    def validate(self, instance, value):
        self._values[instance].validate(value)

class Decorator(metaclass=abc.ABCMeta):
    def __new__(cls, *args, **kwargs):
        if cls.__name__ == "Decorator":
            raise Exception("Decorator is an abstract class. Inherit to implement a concrete decorator.")
        else:
            return super().__new__(cls)

    def __init__(self, **dkwargs):
        super().__init__()
        self.__args = CIStringAttrDict(dkwargs)
        self.__kallable = None

    def __call__(self, kallable):
        self.__kallable = kallable
        return partial(self._decorated, self.__args, self.__kallable)

    @abc.abstractmethod
    def _decorated(self, dargs, kallable, *fvargs, **fkwargs):
        pass

# helpers
def dynamic_load(global_dict, all_list, names):
    for n, b, d in names:
        global_dict[n] = type(n, b, d)
        all_list.append(n)

def prepare_locals(l):
    return {i: j for i, j in l.items() if i not in ['self', '__class__']}


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Arjuna Types

Contains many general purpose type abstractions.
'''

import abc
import pprint
import random
from collections import OrderedDict, namedtuple
from typing import Callable
from enum import Enum, auto

from arjuna.tpi.tracker import track
from arjuna.tpi.constant import DomDirection, DomNodeType

class _ArDict(metaclass=abc.ABCMeta):

    def __init__(self, d=None):
        self.__store = dict()
        self.__key_map = dict()
        if d:
            self.update(d)

    @property
    def orig_dict(self):
        return self.__create_orig_dict()

    @property
    def store(self):
        return self.__store

    def __create_orig_dict(self):
        return {self.__key_map[k]: v for k,v in self.store.items()}

    def __process_key_wrapper(self, key):
        revised_key = self._process_key(key)
        if revised_key not in self.__key_map:
            self.__key_map[revised_key] = key
        return revised_key

    @abc.abstractmethod
    def _process_key(self, key):
        pass

    def __getitem__(self, key):
        return self.__store[self.__process_key_wrapper(key)]

    def pop(self, key):
        return self.__store.pop(self.__process_key_wrapper(key))

    def __setitem__(self, key, value):
        self.__store[self.__process_key_wrapper(key)] = value

    def __delitem__(self, key):
        del self.__store[self.__process_key_wrapper(key)]

    def _update(self, d):
        if not d: return
        for k,v in d.items():
            self[self.__process_key_wrapper(k)] = v

    def update(self, d):
        self._update(d)

    def has_key(self, key):
        return self.__process_key_wrapper(key) in self.__store

    def keys(self):
        return self.__create_orig_dict().keys()

    def __getattr__(self, attr):
        try:
            return self.__store[attr]
        except KeyError:
            raise AttributeError(f"No attribute/key with name {attr}.")

    def __len__(self):
        return len(self.__store.keys())

    def __str__(self):
        if not self.__store:
            return "<empty>"
        else:
            return str(self.__create_orig_dict())

    def __iter__(self):
        return iter(self.__create_orig_dict())

    def clone(self):
        return self._clone()

    def items(self):
        return self.__create_orig_dict().items()

    def is_empty(self):
        return len(self.__store) == 0

    def _get_orig_dict(self):
        return self.__create_orig_dict()


class CIStringDict(_ArDict):
    '''
        Dictionary with case-insensitive keys.

        Arguments:
            pydict: (Optional) A **dict** object.
    '''
    def __init__(self, pydict={}):
        super().__init__(pydict)

    def _process_key(self, key):
        return key.lower()

    def _clone(self):
        return CIStringDict(self._get_orig_dict())

    def __str__(self):
        return "CIStringDict: " + super().__str__()
        

class ProcessedKeyDict(_ArDict):
    '''
        Dictionary with case-insensitive keys.

        Arguments:
            processor: A callable for processing the dictionary key.
            pydict: (Optional) A **dict** object.
    '''

    def __init__(self, *, processor: Callable, pydict: dict={}):
        self.__processor = processor
        super().__init__(pydict)

    def _process_key(self, key):
        return self.__processor(key)  

    def _clone(self):
        return ProcessedKeyDict(self.__processor, self.__store)  

@track("trace")
class OnceOnlyKeyCIStringDict(CIStringDict):
    '''
        Dictionary with case-insensitive keys that allows for immutable key-value pairs.

        Arguments:
            processor: A callable for processing the dictionary key.
            pydict: (Optional) A **dict** object.
    '''

    def __init__(self, d={}):
        super().__init__(d)

    def __setitem__(self, key, value):
        if self.has_key(key):
            raise Exception("You can not change the value once set.")
        super().__setitem__(key, value)

    def _update(self, d):
        if not d: return
        as_dict = dict(d)
        for k in as_dict:
            self[k] = as_dict.get(k)

    def __iter__(self):
        return super().__iter__()

    def _clone(self):
        return OnceOnlyKeyCIStringDict(self.items())


@track("trace")
class Dictable(metaclass=abc.ABCMeta):
    '''
        Abstract class. Any object which has a method **as_dict** is a **Dictable**.
    '''

    @abc.abstractmethod
    def _as_dict(self):
        pass

    def as_dict(self) -> dict:
        '''
            Dictionary representation of this object.

            Returns:
                A **dict** object.
        '''
        retval = self._as_dict()
        if type(retval) is not dict:
            raise Exception("_as_dict must return a dict type. Got {} of type {}".format(retval, type(retval)))
        return retval

class Point(Dictable):
    '''
        Represents an XY coordinate.

        Args:
            x: X co-ordinate
            y: Y co-ordinate
    '''

    def __init__(self, x, y):
        self.__x = x
        self.__y = y

    @property
    def location(self):
        '''
            Get XY co-ordinates as a tuple -> (x,y)
        '''
        return (self.__x, self.__y)

    def _as_dict(self):
        return {"x": self.__x, "y":self.__y}

class Offset(Point):
    '''
        Represents an offset from current **Point** on **Screen** in terms of XY coordinates.

        Args:
            x: X co-ordinate
            y: Y co-ordinate
    '''

    def __init__(self, x, y):
        super().__init__(x,y)

class oneof:
    '''
        Represents the given sequence as choices.

        Args:
            choices: Arbitrary objects as choices.
    '''

    def __init__(self, *choices):
        self.__choices = choices

    def as_list(self):
        return list(self.__choices)

@track("trace")
class Screen:
    '''
        Represents Gui Screen in terms of its XY coordinates.
    '''

    @staticmethod
    def xy(x:int, y:int) -> Point:
        '''
            Create a **Point** on Gui Screen in terms of its XY coordinates.
        '''
        #return  _Point(x,y)
        raise NotImplementedError()

    @staticmethod
    def offset(x, y) -> Offset:
        '''
            Create a **Offset** on Gui Screen in terms of its XY coordinates.
        '''
        # return _Offset(x,y)
        raise NotImplementedError()

@track("trace")
class nvpair(Dictable):
    '''
        Encapsulates a name-value pair. It is an implementation of **Dictable**.

        Args:
            name: Name of this object
            value: Value of this object
    '''

    def __init__(self, name, value):
        self.__name = name
        self.__value = value

    @property
    def name(self):
        '''
        Name of this object.
        '''
        return self.__name

    @property
    def value(self):
        '''
        Value of this object.
        '''
        return self.__value

    def _as_dict(self):
        return {"name" : self.__name, "value": self.__value}


@track("trace")
class attr(nvpair):
    '''
        A name-value pair with an associated optional tag name. It is an implementation of **Dictable**.

        Keyword Arguments:
            attr_name_value: (Mandatory) Key-Value pair representing name and value of attribute.

        Note:
            In case the attribute name conflicts with a Python language keyword, you can prefix it with '__' i.e. two underscores. These underscores are removed while processing the attribute name.

            For example, '__for' will become 'for' to avoid conflict with Python's **for** keyword.
    '''

    def __init__(self, **attr_name_value):
        if len(attr_name_value) > 1:
            raise Exception("attr specification must contain a single key value pair for attribute name and value")
        if len(attr_name_value) > 1:
            raise Exception("attr/fattr/battr/eattr specification should contain only a single key value pair for attribute name and value")
        name = list(attr_name_value.keys())[0]
        if name.startswith('__'):
            name = name[2:]
        value = list(attr_name_value.values())[0]
        super().__init__(name, value)

    def _as_dict(self):
        d = super()._as_dict()
        return d


@track("trace")
class nvpairs(Dictable):
    '''
        Encapsulates arbitrary name-value pairs. It is an implementation of **Dictable**.

        Keyword Arguments:
            **nvpairs: Arbitrary name-value pairs passed as keyword arguments.
    '''
    def __init__(self, **nvpairs):
        self.__kwargs = nvpairs

    def _as_dict(self):
        return self.__kwargs


@track("trace")
class withx(nvpairs):
    '''
        Encapsulates arbitrary name-value pairs. It is an implementation of **Dictable**.

        This is to be used when using withx locators progammaitcally.

        Keyword Arguments:
            **nvpairs: Arbitrary name-value pairs passed as keyword arguments.
    '''
    pass

@track("trace")
class Node(nvpairs):

    def __init__(self, *, type, **nvpairs):
        '''
            Represents a Node in HTML/XML/DOM described by name-value pairs.

            Keyword Arguments:
                type: (Mandatory) DomNodeType - NODE/FNODE/BNODE determines how attribute content is matched:
                    * NODE: Partial Content
                    * FNODE: Full Content
                    * BNODE: Partial Content at Beginning
                **nvpairs: Special and Arbitrary name-value pairs passed as keyword arguments.

            Note:
                Following keywords have special meaning:

                    * tag - Represents tag name of node
                    * classes - Passed as a string or list/tuple represents one or more classes in class attribute of node (order does not matter)
                    * text - Represents text content
                    * star_text - It is equivalent of *//text() in XPath
                    * dot_text - It is equivalent of . in XPath
                    * attrs - It is a dictionary of attributes. Can be used when names can not be passed as keywords.
                        * Names conflict with Python keywords. For example: 'for'
                        * '.text' for dot_text
                        * '*text' for star_text

                All other key value pairs are assumed to be attribute names and corresponding values.

                If any keyword is preceded with '__' (double underscores), the underscores are removed at the time of definition generation. This can be used to avoid conflict of attribute names with Python keywords.
        '''
        super().__init__(**nvpairs)
        self.__type = type
    
    @property
    def ntype(self):
        return self.__type

    def _as_dict(self):
        out_dict = dict()
        d = super()._as_dict()
        if 'attrs' in d:
            for k,v in d['attrs'].items():
                if k.startswith('__'):
                    k = k[2:]
                out_dict[k] = v
        for k,v in d.items():
            if k == 'attrs': continue
            if k.startswith('__'):
                k = k[2:]
            out_dict[k] = v
        return out_dict

@track("trace")
class node(Node):

    def __init__(self, **nvpairs):
        super().__init__(type=DomNodeType.NODE, **nvpairs)

@track("trace")
class fnode(Node):

    def __init__(self, **nvpairs):
        super().__init__(type=DomNodeType.FNODE, **nvpairs)

@track("trace")
class bnode(Node):

    def __init__(self, **nvpairs):
        super().__init__(type=DomNodeType.BNODE, **nvpairs)

@track("trace")
class _Axis:

    def __init__(self, direction, node):
        self.__direction = direction
        self.__node = node

    @property
    def direction(self):
        return self.__direction

    @property
    def node(self):
        return self.__node


@track("trace")
class axes:
    '''
        Represents a DOM axes with a starting node and then as per axis objects sequence.

        Arguments:
            start: Starting point as a node object
    '''

    def __init__(self, start):
        self.__start = start
        self.__axes = list()

    @property
    def _start(self):
        return self.__start

    @property
    def _axes(self):
        return self.__axes

    @classmethod
    def _from_dict(cls, axes_dict):
        axes_dict = {k.lower():v for k,v in axes_dict.items()}
        axes_obj = axes(node(**axes_dict['start']))
        del axes_dict['start']
        for k,v in axes_dict.items():
            getattr(axes_obj, k)(node(**v))
        return axes_obj

    @classmethod
    def _from_list(cls, axes_list):
        start = None
        axes_entries = []

        for entry in axes_list:
            if len(entry) > 1:
                raise Exception("If axes specification is of list type, each list item must be a single key value pair. Wrong axes definition found: {}".format(axes_list))
            name = list(entry.keys())[0].lower()
            value = list(entry.values())[0]
            if name == "start":
                if start is None:
                    start = node(**value)
                else:
                    raise Exception("Duplicate definition of start key found in axes def: {}".format(axes_list))
            else:
                axes_entries.append((name, node(**value)))
        
        if start is None:
            raise Exception("axes definition must have start key. Wrong axes def: {}".format(axes_list))

        axes_obj = axes(start)
        for direction, n in axes_entries:
            getattr(axes_obj, direction)(n)
        return axes_obj

    def up(self, node):
        '''
        Move towards node definition in the direction of Ancestors.
        '''
        self.__axes.append(_Axis(DomDirection.UP, node))
        return self

    def down(self, node):
        '''
        Move towards node definition in the direction of Descendants.
        '''
        self.__axes.append(_Axis(DomDirection.DOWN, node))
        return self

    def left(self, node):
        '''
        Move towards node definition in the direction of Previous Siblings.
        '''
        self.__axes.append(_Axis(DomDirection.LEFT, node))
        return self

    def right(self, node):
        '''
        Move towards node definition in the direction of Forward Siblings.
        '''
        self.__axes.append(_Axis(DomDirection.RIGHT, node))
        return self

class NotFound:
    '''
    To differentiate a not found object from Python's None.

    Always evalutes to False in a boolean context.
    '''
    
    def __bool__(self):
        return False

NetworkPacketInfo = namedtuple("NetworkPacketInfo", "label request response sub_network_packets")


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from enum import Enum

from arjuna.configure.options import EditableConfig
from arjuna.core.value import Value
from arjuna.core.utils import file_utils
from arjuna.core.constant import ConfigStage


class TestConfigurator:
    _EMPTY_CONF = EditableConfig.empty_conf()

    def __init__(self, root_dir, cli_config, run_id, linked_projects=[]):
        self.__root_dir = root_dir
        self.__cli_config = cli_config
        self.__run_id = run_id
        self.__default_ref_config = None
        self.__data_confs = None
        self.__env_confs = None
        self.__dataconf_enconf_confs = dict()
        self.__project_config_loaded = False
        self.__linked_projects = linked_projects

        self.__load()

    def __load(self):
        self.__load_central_conf()
        self.__load_project_conf()
        self.__load_cli_dicts()
        self.__load_data_confs()
        self.__load_env_confs()

        self.__default_ref_config.update(self.__data_confs["data"])
        self.__default_ref_config.update(self.__env_confs["env"])
        self.__update_config(self.__default_ref_config)
        # self.__process_data_configs()

        self.__load_combinations()

    def __load_central_conf(self):
        self.__default_ref_config = EditableConfig.arjuna_conf(project_root_dir=self.__root_dir, run_id=self.__run_id)

    def __load_project_conf(self):
        self.__default_ref_config = EditableConfig.project_conf(arjuna_conf=self.__default_ref_config, linked_projects=self.__linked_projects)

    def __load_data_confs(self):
        self.__data_confs = EditableConfig.data_confs(arjuna_conf=self.__default_ref_config) 
        if "data" not in self.__data_confs:
            self.__data_confs["data"] = self._EMPTY_CONF

    def __load_env_confs(self):
        self.__env_confs = EditableConfig.env_confs(arjuna_conf=self.__default_ref_config)
        if "env" not in self.__env_confs:
            self.__env_confs["env"] = self._EMPTY_CONF

    def load_options_from_file(self, fpath, *, conf_stage):
        if file_utils.is_absolute_path(fpath):
            if not file_utils.is_file(fpath):
                if file_utils.is_dir(fpath):
                    raise Exception("Not a file: {}".format(fpath))
                else:
                    raise Exception("File does not exist: {}".format(fpath))
        else:
            from arjuna import Arjuna, ArjunaOption
            conf_dir = self.__default_ref_config.arjuna_options.value(ArjunaOption.CONF_DIR)
            fpath = os.path.abspath(os.path.join(conf_dir, fpath))
            if not file_utils.is_file(fpath):
                if file_utils.is_dir(fpath):
                    raise Exception("Not a file: {}".format(fpath))
                else:
                    raise Exception("File does not exist: {}".format(fpath))
        return EditableConfig.from_file(file_path=fpath, creation_context=f"Ad-hoc configuration file at {fpath}", conf_stage=conf_stage)

    def __load_cli_dicts(self):
        self.__cli_config = EditableConfig.from_maps(ref_config=None, arjuna_options=self.__cli_config.arjuna_options, user_options=self.__cli_config.user_options, conf_stage=ConfigStage.CLI)

    def __update_config_for_env(self, config):
        if self.__chosen_env_conf:
            config.update(self.__chosen_env_conf)

    def __update_config_for_data_conf(self, config):
        if self.__data_conf:
            config.update(self.__data_conf)

    def __update_config_for_cli(self, config):
        config.update(self.__cli_config)

    def __update_config(self, config):
        self.__update_config_for_cli(config)
        config.process_arjuna_options()

    def __update_ref_config(self):
        self.__update_config(self.__default_ref_config, update_from_chosen_env=True)

    def __load_one_combo(self, dconf, econf):
        conf = EditableConfig.empty_conf()
        conf.update(self.__default_ref_config)
        conf.update(dconf)
        conf.update(econf)
        self.__update_config(conf)
        return conf

    def __load_combinations(self):
        def __add(name, conf):
            if name not in self.__dataconf_enconf_confs:
                self.__dataconf_enconf_confs[name] = EditableConfig.empty_conf()
            self.__dataconf_enconf_confs[name].update(conf)

        # Load combinations from linked projects
        for linked_project in self.__linked_projects:
            confs = linked_project.data_env_conf_map
            for name, conf in confs.items():
                __add(name, conf)

        for dname, dconf in self.__data_confs.items():
            conf = self.__load_one_combo(dconf, self._EMPTY_CONF)
            __add(dname, conf)
            
        for ename, econf in self.__env_confs.items():
            conf = self.__load_one_combo(self._EMPTY_CONF, econf)
            __add(ename, conf)

        for dname, dconf in self.__data_confs.items():
            for ename, econf in self.__env_confs.items():
                cname = "{}_{}".format(dname, ename)
                conf = self.__load_one_combo(dconf, econf)
                __add(cname, conf)

    @property
    def ref_config(self):
        return self.__default_ref_config

    @property
    def file_confs(self):
        return self.__dataconf_enconf_confs

    def register_new_config(self, arjuna_options, user_options, *, conf_stage, parent_config=None):
        # Registering a config is post project conf registration. If no project conf, set it to true.
        self.__project_config_loaded = True
        reference = parent_config and parent_config._wrapped_config or self.__default_ref_config._wrapped_config
        conf = EditableConfig.from_maps(ref_config=reference, conf_stage=conf_stage, arjuna_options=arjuna_options, user_options=user_options)
        self.__update_config(conf)
        return conf


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, auto

class CommandEnum(Enum):
    CREATE_PROJECT = auto()
    RUN_PROJECT = auto()
    RUN_SESSION = auto()
    RUN_STAGE = auto()
    RUN_GROUP = auto()
    RUN_SELECTED = auto()

class TargetEnum(Enum):
    PROJECT = auto()
    GROUP = auto()
    SELECTED = auto()


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from .internal.repr.request import HttpRequestYamlRepr
from .internal.repr.response import HttpResponseYamlRepr
from arjuna.tpi.helper.arjtype import CIStringDict

class HttpMessage:

    def __init__(self, session, req_repr, resp_processor):
        self.__session = session
        self.__req_repr = req_repr
        self.__resp_processor = resp_processor
        self.__label = req_repr.label

    @property
    def label(self):
        return self.__label

    @property
    def _req(self):
        return self.__req_repr

    @property
    def _session(self):
        return self.__session

    def send(self):
        method = self._req.method
        try:
            call = getattr(self._session, method)
        except AttributeError:
            raise Exception(f"Unsupported HTTP method: {method}")
        else:
            response = call(self._req.route, label=self.label, **self._req.attrs)
            self.__resp_processor.validate(response)
            return response

    @classmethod
    def root(cls, session):
        req_repr = HttpRequestYamlRepr(session, CIStringDict(), label="Root")
        resp_proc = HttpResponseYamlRepr(session, CIStringDict())
        return HttpMessage(session, req_repr, resp_proc)

    @classmethod
    def from_file(cls, session, msg_file_name, **fargs):
        # Process Yaml file
        from arjuna import C, Yaml
        file_path = os.path.join(C("httpauto.message.dir"), msg_file_name + ".yaml")
        f = open(file_path, "r")
        msg_yaml = f.read()
        f.close()
        from arjuna.core.fmt import arj_format_str, arj_convert
        # Convert Arjuna custom objects to raw Python objects before formatting.
        fargs = {k:arj_convert(v) for k,v in fargs.items()}
        msg_yaml = arj_format_str(msg_yaml, tuple(), fargs)
        msg_yaml = Yaml.from_str(msg_yaml, allow_any=True)
        if msg_yaml is None:
            return cls.root(session)

        if "label" in msg_yaml:
            label = msg_yaml["label"]
            del msg_yaml["label"]
        else: 
            label = msg_file_name

        if "request" in msg_yaml:
            req_repr = HttpRequestYamlRepr(session, CIStringDict(msg_yaml["request"].as_map()), label=label)
            del msg_yaml["request"]
        else:
            req_repr = HttpRequestYamlRepr(session, CIStringDict(), label=label)
        
        resp_proc = HttpResponseYamlRepr(session, CIStringDict(msg_yaml.as_map()))

        return HttpMessage(session, req_repr, resp_proc)




# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import xlrd


class ExcelRowReader:
    def __init__(self, path):
        self.wb = xlrd.open_workbook(path)
        self.sheet = self.wb.sheet_by_name(self.wb.sheet_names()[0])
        self.rcount = self.sheet.nrows
        self.ccount = self.sheet.ncols
        self.curent_row_index = -1
        self.validate()

    def __iter__(self):
        return self

    def read_next_row(self):
        self.curent_row_index += 1
        if self.curent_row_index < self.rcount:
            return self.sheet.row(self.curent_row_index)
        else:
            raise Exception("Done")

    def next(self):
        try:
            return self.process(self.read_next_row())
        except:
            self.close()
            raise StopIteration()

    __next__ = next

    def read(self):
        return [r for r in iter(self)]

    def process(self, row):
        return row

    def validate(self):
        pass

    def close(self):
        pass
        #self.wb.close()


class ExcelRow2ArrayReader(ExcelRowReader):
    def __init__(self, path):
        super().__init__(path)
        self.__headers = []
        self._populate_headers()

    def get_headers(self):
        return self.__headers

    @property
    def headers(self):
        return self.__headers

    def _populate_headers(self):
        self.__headers = [h.value for h in self.read_next_row()]

    def process(self, row):
        return [h.value for h in row]

    def validate(self):
        if self.rcount == 0:
            raise Exception("Empty or wrongly formattted Excel file. Is first line empty?")

    def __iter__(self):
        return self

    def next(self):
        return super().next()

class ExcelRow2MapReader(ExcelRow2ArrayReader):
    def __init__(self, path):
        super().__init__(path)

    def process(self, row):
        return dict(zip(self.headers, super().process(row)))



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.guiauto.source.multielement import GuiMultiElementSource
from arjuna.tpi.tracker import track

@track("debug")
class GuiRadioGroup:
    '''
        Represents a radio group in th Gui.

        Not meant to be directly created. It is created using calls from **Gui** object or **GuiNamespace** object of **Gui**.

        Arguments:
            gui: **Gui** object containing this GuiDropDown.
            wmd: **GuiElementMetaData** object for this GuiDropDown.

        Keyword Arguments:
            parent: **GuiElement** in case it is found inside a **GuiElement**. Default is the **Gui** object.
    '''

    def __init__(self, gui, wmd, parent=None):
        self.__gui = gui
        self.__wmd = wmd
        self.__automator = gui._automator
        self.__finder = parent and parent or gui
        self.__radios = None
        self.__find(self.__finder, wmd)

    @property
    def gui(self) -> 'Gui':
        '''
            **Gui** object containing this GuiDropDown.
        '''
        return self.__gui

    def __validate_radio_buttons(self, source):
        if [t for t in source.tag_names if t.strip().lower() != 'input']:
            raise Exception("Not a valid radio group. Contains non-input elements.")
        if [t for t in source.get_attr_values("type") if t.strip().lower() != 'radio']:
            raise Exception("Not a valid radio group. Contains non-radio elements.")
        names = source.get_attr_values("name")
        if len(set(names)) != 1:
            raise Exception("Not a valid radio group. Contains radio elements belonging to different radio groups.")

    def __check_type_if_configured(self, tags):
        if self.__wmd.meta.settings.should_check_type(): self.__validate_radio_buttons(tags)

    def __find(self, finder, wmd):
        # This would force the identification of partial elements in the wrapped multi-element.
        self.__radios = finder._wmd_finder.multi_element(wmd)
        self.__check_type_if_configured(self.source)

    def has_index_selected(self, index: int) -> bool:
        '''
            Check if this GuiRadioGroup has radio button selected at given index .

            Args:
                index: Target index.
        '''
        return self.__radios[index].is_selected()

    def has_value_selected(self, value: str) -> bool:
        '''
            Check if this GruiRadioGroup has radio button with given value attribute content selected.

            Args:
                value: Exact content of value attribute.
        '''
        return self.__radios.get_element_by_value(value).is_selected()

    @property
    def value(self) -> str:
        '''
            Content of value attribute of selected radio button.
        '''
        instance = self.__radios.first_selected_element
        return instance.source.get_attr_value("value")

    def __select_option(self, option):
        option.select()
        if self.__wmd.meta.settings.should_check_post_state() and not option.is_selected():
            raise Exception("The attempt to select the radio button was not successful.")

    def select_index(self, index: int) -> None:
        '''
            Select radio button at given index.

            Args:
                index: Target index.
        '''
        option = self.__radios[index]
        self.__select_option(option)

    def select_ordinal(self, ordinal: int) -> None:
        '''
            Select radio group at given ordinal.

            Ordinals are as per human counting. First element is at ordinal 1.

            Args:
                ordinal: Target ordinal.
        '''
        return self.select_by_index(ordinal-1)

    def select_value(self, value: str) -> None:
        '''
            Select radio button with given content of value attribute.

            Args:
                value: Exact content of value attribute.
        '''
        option = self.__radios.get_element_by_value(value)
        self.__select_option(option)

    @property
    def source(self) -> GuiMultiElementSource:
        '''
            **GuiMultiElementSource** for this GuiRadioGroup (source of associated GuiMultiElement)
        '''
        return self.__radios.source


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.constant import ArjunaOption

from arjuna.core.poller.conditions import *
from arjuna.core.poller.caller import *
from arjuna.tpi.tracker import track

class BrowserConditions:

    def __init__(self, browser):
        self.__browser = browser

    @property
    def gui(self):
        return self.__gui

    def DocumentReadyState(self):
        caller = DynamicCaller(self.__browser.is_document_ready)
        return BooleanCondition(caller)

@track("info")
class Browser:

    def __init__(self, automator):
        self.__automator = automator
        self.__conditions = BrowserConditions(self)

    def is_document_ready(self):
        return self.execute_javascript("return document.readyState") == "complete"

    def go_to_url(self, url):
        self.__automator.dispatcher.go_to_url(url=url)
        self.__conditions.DocumentReadyState().wait()

    def go_back(self):
        self.__automator.dispatcher.go_back_in_browser()
        self.__conditions.DocumentReadyState().wait()

    def go_forward(self):
        self.__automator.dispatcher.go_forward_in_browser()
        self.__conditions.DocumentReadyState().wait()

    def refresh(self, hard=False):
        if hard:
            self.__automator.execute_javascript("window.location.reload(true);")
        else:
            self.__automator.execute_javascript("window.location.reload(false);")
        self.__conditions.DocumentReadyState().wait()

    def execute_javascript(self, js, *args):
        return self.__automator.dispatcher.execute_javascript(js, *args)


import re
from arjuna.core.constant import GuiElementType
from arjuna.tpi.constant import DomNodeType, DomDirection

from ._with import GuiGenericLocator
from .enums import GenericLocateWith

class LocatorTranslator:
    '''
        Translates With Constructs to what underlying engines understand.
    '''

    BASIC_LOCATORS = {
        GenericLocateWith.ID,
        GenericLocateWith.NAME,
        GenericLocateWith.XPATH,
        GenericLocateWith.IMAGE,
        GenericLocateWith.INDEX,
        GenericLocateWith.WINDOW_TITLE,
        GenericLocateWith.WINDOW_PTITLE,
        GenericLocateWith.JS,
    }

    NEED_TRANSLATION = {
        GenericLocateWith.LINK : GenericLocateWith.PARTIAL_LINK_TEXT,
        GenericLocateWith.FLINK : GenericLocateWith.LINK_TEXT,
        GenericLocateWith.SELECTOR : GenericLocateWith.CSS_SELECTOR,
    }

    XTYPE_LOCATORS = {
        GuiElementType.TEXTBOX: "//input[@type='text']",
        GuiElementType.PASSWORD: "//input[@type='password']",
        GuiElementType.LINK: "//a",
        GuiElementType.BUTTON: "//input[@type='button']",
        GuiElementType.SUBMIT_BUTTON: "//input[@type='submit']",
        GuiElementType.DROPDOWN: "//select",
        GuiElementType.CHECKBOX: "//input[@type='checkbox']",
        GuiElementType.RADIO: "//input[@type='radio']",
        GuiElementType.IMAGE: "//img",
    }

    XPATH_LOCATORS = {
        GenericLocateWith.TEXT : "//*[contains(text(),'{}')]",
        GenericLocateWith.FTEXT : "//*[text()='{}']",
        GenericLocateWith.BTEXT : "//*[starts-with(text(),'{}')]",
        GenericLocateWith.VALUE : "//*[@value='{}']",
        GenericLocateWith.TITLE : "//*[@title='{}']",
        GenericLocateWith.IMAGE_SRC : "//img[@src='{}']"
    }

    NAMED_ARG_LOCATORS = {
        # GenericLocateWith.ATTR : "//*[contains(@{name},'{value}')]",
        # GenericLocateWith.FATTR : "//*[@{name}='{value}']",
        # GenericLocateWith.BATTR : "//*[starts-with(@{name},'{value}')]",
        # GenericLocateWith.EATTR : "//*[ends-with(@{name},'{value}')]",
        GenericLocateWith.ATTR : "*[{name}*='{value}']",
        GenericLocateWith.FATTR : "*[{name}='{value}']",
        GenericLocateWith.BATTR : "*[{name}^='{value}']",
        GenericLocateWith.EATTR : "*[{name}$='{value}']",
        GenericLocateWith.POINT : "return document.elementFromPoint({x}, {y})",
    }

    TEXT_TRANSLATIONS = {
        'text' : "text()",
        '*text' : "*//text()",
        'star_text' : "*//text()",
        '.text' : ".",
        'dot_text' : ".",
    }

    @classmethod
    def __process_selector(cls, cssvalue):
        css_string = None
        if type(cssvalue) is str:
            css_string = "." + cssvalue.replace('.', ' ').strip()
        else:
            if type(cssvalue[0]) is str:
                # Tuple like ('button button-large',)
                cssvalue = " ".join(cssvalue)
                css_string = "." + cssvalue.replace('.', ' ').strip()
            else:
                # Tuple is (('button', 'button-large'),)
                css_string = "." + ".".join(cssvalue[0])
        return re.sub(r'\s+', '.', css_string)

    @classmethod
    def __convert_to_class_list(cls, cssvalue):
        return cls.__process_selector(cssvalue)[1:].split(".")

    @classmethod
    def __process_tags(cls, tagsvalue):
        tag_list = None
        if type(tagsvalue) is str:
            tag_list = tagsvalue.strip().split()
        else:
            tag_list = tagsvalue
        return [t.lower()=='any' and '*' or t for t in tag_list]

    @classmethod
    def translate(cls, locator):
        from arjuna import log_debug
        rltype = locator.ltype
        rlvalue = locator.lvalue
        glvalue = None
        try:
            gltype = GenericLocateWith[rltype.upper()]
        except Exception as e:
            raise Exception("Invalid locator across all automators: {}={}. Error: {}".format(rltype, type(rlvalue), str(e)))
        else:
            log_debug("Processing locator: Type: {}; Value: {}".format(str(gltype), rlvalue))
            if gltype == GenericLocateWith.ELEMENT:
                glvalue = rlvalue
            elif gltype in cls.BASIC_LOCATORS:
                glvalue = rlvalue
            elif gltype in cls.NEED_TRANSLATION:
                glvalue = rlvalue
                gltype = cls.NEED_TRANSLATION[gltype]
            elif gltype in cls.XPATH_LOCATORS:
                glvalue = cls.XPATH_LOCATORS[gltype].format(rlvalue)
                gltype = GenericLocateWith.XPATH
            elif gltype == GenericLocateWith.POINT:
                glvalue = cls.NAMED_ARG_LOCATORS[gltype].format(**rlvalue)
            elif gltype in cls.NAMED_ARG_LOCATORS:
                glvalue = cls.NAMED_ARG_LOCATORS[gltype].format(**rlvalue)
                gltype = GenericLocateWith.CSS_SELECTOR
            elif gltype == GenericLocateWith.CLASSES:
                css_string = cls.__process_selector(rlvalue)
                glvalue = cls.__process_selector(rlvalue)
                gltype = GenericLocateWith.CSS_SELECTOR
            elif gltype == GenericLocateWith.TAGS:
                glvalue = " ".join(cls.__process_tags(rlvalue))
                gltype = GenericLocateWith.CSS_SELECTOR                
            elif gltype in {GenericLocateWith.NODE, GenericLocateWith.BNODE, GenericLocateWith.FNODE}:
                gltype, glvalue = cls.__translate_node(gltype, rlvalue)
            elif gltype == GenericLocateWith.AXES:
                gltype, glvalue = cls.__translate_axes(gltype, rlvalue)
            else:
                raise Exception("Locator not supported yet by Arjuna: " + rltype)    
        return GuiGenericLocator(gltype, glvalue)


    @classmethod
    def __translate_node_rlvalue_to_xpath(cls, gltype, rlvalue, prefix_slashes=True):
        xblocks = []
        tags = '*'
        for k,v in rlvalue.items():
            if k.lower() == 'tags':
                tags = "//".join(v)
                continue

            if k == "classes":
                for c in v:
                    xblocks.append(f"contains(@class,'{c}')")
                continue

            if k.lower() in cls.TEXT_TRANSLATIONS:
                k = cls.TEXT_TRANSLATIONS[k.lower()]
            else:
                k = f'@{k}'
            if gltype == GenericLocateWith.NODE:
                xblocks.append(f"contains({k},'{v}')")
            elif gltype == GenericLocateWith.BNODE:
                xblocks.append(f"starts-with({k},'{v}')")
            elif gltype == GenericLocateWith.FNODE:
                xblocks.append(f"{k}='{v}'")
        if xblocks:
            xblocks_str = "[" + " and ".join(xblocks) + "]"
        else:
            xblocks_str = ""
        
        if prefix_slashes:
            prefix = "//"
        else:
            prefix = ""
        glvalue = f"{prefix}{tags}{xblocks_str}"
        gltype = GenericLocateWith.XPATH
        return gltype, glvalue

    @classmethod
    def __translate_node_to_selector(cls, gltype, rlvalue):
        cblocks = []
        tags = '*'
        classes = ""
        for k,v in rlvalue.items():
            if k.lower() == 'tags':
                tags = " ".join(v)
                continue

            if k == "classes":
                classes = "." + ".".join(v)
                continue
            
            if gltype == GenericLocateWith.NODE:
                cblocks.append(f"[{k}*='{v}']")
            elif gltype == GenericLocateWith.BNODE:
                cblocks.append(f"[{k}^='{v}']")
            elif gltype == GenericLocateWith.FNODE:
                cblocks.append(f"[{k}='{v}']")
        if cblocks:
            cblocks_str = "".join(cblocks)
        else:
            cblocks_str = ""
        glvalue = f"{tags}{classes}{cblocks_str}"
        gltype = GenericLocateWith.CSS_SELECTOR
        return gltype, glvalue


    @classmethod
    def __process_node_rlvalue_and_xpath_condition(cls, rlvalue):
        lkeys = [k.lower() for k in rlvalue.keys()]
        if 'tag' in lkeys and 'tags' in lkeys:
            raise Exception("node Locator definition can contain either 'tag' or 'tags' as key.")
        if len(set(lkeys).intersection(set(cls.TEXT_TRANSLATIONS.keys()))) > 1:
            raise Exception("node Locator definition can contain only one of text/star_text/*text/dot_text/.text keys.")
        # Initial processing
        updated_rlvalue = dict()
        use_xpath = False
        for k,v in rlvalue.items():
            if k.lower() == 'tag':
                updated_rlvalue['tags'] = cls.__process_tags(v)
            elif k.lower() == 'tags':
                updated_rlvalue['tags'] = cls.__process_tags(v)
            elif k.lower() == 'classes':
                updated_rlvalue['classes'] = cls.__convert_to_class_list(v)
            elif k.lower() in cls.TEXT_TRANSLATIONS:
                updated_rlvalue[k.lower()] = v
                use_xpath = True
            elif k.lower() == "use_xpath":
                from arjuna.core.types.constants import TRUES, FALSES
                v = str(v).strip()
                if v.lower() not in TRUES and v.lower() not in FALSES:
                    raise Exception("Allowed values for use_path are true/false/on/off")
                if v.lower() in TRUES:
                    use_xpath = True
            else:
                updated_rlvalue[k] = v
        return updated_rlvalue, use_xpath

    @classmethod
    def __translate_node(cls, gltype, rlvalue):
        rlvalue, use_xpath = cls.__process_node_rlvalue_and_xpath_condition(rlvalue)

        if use_xpath:
            return cls.__translate_node_rlvalue_to_xpath(gltype, rlvalue)
        else:
            return cls.__translate_node_to_selector(gltype, rlvalue)

    _NTYPE_GLTYPE_MAP ={
        DomNodeType.NODE: GenericLocateWith.NODE,
        DomNodeType.FNODE: GenericLocateWith.FNODE,
        DomNodeType.BNODE: GenericLocateWith.BNODE,
    }

    _DIRECTION_TO_XPATH_REL_MAP ={
        DomDirection.UP: "/ancestor::",
        DomDirection.DOWN: "/descendant::",
        DomDirection.LEFT: "/preceding-sibling::",
        DomDirection.RIGHT: "/following-sibling::"
    }

    @classmethod
    def __translate_axes(cls, gltype, rlvalue):    
        from arjuna.tpi.helper.arjtype import Node, axes
        def create_xpath_for_part(gltype, locator_value, *, prefix_slashes=False):
            rlvalue, use_xpath = cls.__process_node_rlvalue_and_xpath_condition(locator_value)
            return cls.__translate_node_rlvalue_to_xpath(gltype, rlvalue, prefix_slashes=prefix_slashes)

        xpath_parts = []

        # Process Starting point
        if type(rlvalue) is dict:
            rlvalue = axes._from_dict(rlvalue)
        elif type(rlvalue) is list:
            rlvalue = axes._from_list(rlvalue)
        start = rlvalue._start
        start_gltype = cls._NTYPE_GLTYPE_MAP[start.ntype]
        _, start_rlvalue = create_xpath_for_part(start_gltype, start.as_dict(), prefix_slashes=True)
        xpath_parts.append(start_rlvalue)

        for axis in rlvalue._axes:
            xpath_parts.append(cls._DIRECTION_TO_XPATH_REL_MAP[axis.direction])
            axis_gltype = cls._NTYPE_GLTYPE_MAP[axis.node.ntype]
            _, axis_rlvalue = create_xpath_for_part(axis_gltype, axis.node.as_dict())
            xpath_parts.append(axis_rlvalue)

        return GenericLocateWith.XPATH, "".join(xpath_parts)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import os
import time
import datetime
import json

from arjuna.tpi.constant import ArjunaOption
from arjuna.tpi.guiauto.base.container import GuiWidgetContainer
from arjuna.interact.gui.auto.base.dispatchable import _Dispatchable
from .drivercaps import DriverCapabilities
from arjuna.tpi.guiauto.source.page import GuiPageSource
from arjuna.interact.gui.dispatcher.selenium.driver import SeleniumDriverDispatcher
from arjuna.interact.gui.auto.finder.wmd import GuiWidgetMetaData
from arjuna.tpi.guiauto.widget.element import GuiElement
from arjuna.tpi.helper.image import Image
from arjuna.interact.gui.auto.automator.network_recorder import BrowserMobNetworkRecorder
from arjuna import track, log_debug

@track("debug")
class GuiAutomator(GuiWidgetContainer,_Dispatchable):

    def __init__(self, app, config, ext_config=None):
        GuiWidgetContainer.__init__(self, config)
        _Dispatchable.__init__(self)
        self.__app = app
        self.__econfig = ext_config
        self.__create_screenshots_dir()
        self.__main_window = None
        self.__in_slomo = config.value(ArjunaOption.GUIAUTO_SLOMO_ON)
        self.__slomo_interval = config.value(ArjunaOption.GUIAUTO_SLOMO_INTERVAL)

        from .webalert_handler import WebAlertHandler
        from .automator_conditions import GuiAutomatorConditions
        from .viewcontext_handler import ViewContextHandler
        self.__alert_handler = WebAlertHandler(self)
        self.__conditions_handler = GuiAutomatorConditions(self)
        self.__view_handler = ViewContextHandler(self)
        self.__browser = None
        self.__screenshots_dir = config.value(ArjunaOption.SCREENSHOTS_DIR)

        self.__source_parser = None

        # As of now it directly connects to Selenium Dispatcher
        # Code should be introduced here which passes through DispatcherPicker
        # based on choice of engine to support more libs.
        self._dispatcher = SeleniumDriverDispatcher()
        self._network_recorder = BrowserMobNetworkRecorder(self)
        self.__launch()

    @property
    def app(self):
        return self.__app

    @property
    def ext_config(self):
        return self.__econfig

    @property
    def network_recorder(self):
        return self._network_recorder

    @property
    def screenshots_dir(self):
        return self.__screenshots_dir

    def create_wmd(self, *locators):
        return GuiWidgetMetaData.create_wmd(*locators)

    def get_source_from_remote(self):
        return self.dispatcher.get_source()

    # def create_dispatcher(self):
    #     self._set_dispatcher(self.dispatcher_creator.create_gui_automator_dispatcher(self.config, self.setu_id))

    def slomo(self):
        if self.__in_slomo:
            time.sleep(self.__slomo_interval)

    def set_slomo(self, on, interval=None):
        self.__in_slomo = on
        if interval is not None:
            self.__slomo_interval = interval

    @property
    def browser(self):
        return self.__browser

    @property
    def main_window(self):
        return self.__main_window

    def child_window(self, wmd):
        return self.get_main_window().get_child_window(wmd)

    @property
    def latest_child_window(self):
        return self.get_main_window().get_latest_child_window()

    def close_all_child_windows(self):
        self.main_window.close_all_child_windows()

    def get_dom_root(self, gui):
        raise NotImplementedError()
        # from arjuna.tpi.guiauto.widget.frame import DomRoot
        # return DomRoot(gui)

    def get_frame(self, gui, wmd):
        return self.dom_root(gui).frame(wmd)

    @property
    def alert_handler(self):
        return self.__alert_handler

    @property
    def view_handler(self):
        return self.__view_handler

    @property
    def conditions(self):
        return self.__conditions_handler

    def __create_screenshots_dir(self):
        sdir = self.config.value(ArjunaOption.SCREENSHOTS_DIR)
        if not os.path.isdir(sdir):
            os.makedirs(sdir)

    # #Override
    # def _get_object_uri(self):
    #     return self.__automator_uri

    def __launch(self):
        caps = DriverCapabilities(self.config, self.__econfig)
        self.dispatcher.launch(caps.processed_config)

        from arjuna.tpi.guiauto.obj.window import MainWindow
        self.__main_window = MainWindow(self.app, self)

        from .browser import Browser
        self.__browser = Browser(self)

    def quit(self):
        self.dispatcher.quit()

    def __screenshot(self, file_path):
        # switch_view_context = None
        # if self.config.value(ArjunaOption.MOBILE_OS_NAME).lower() == "android":
        #     view_name = self.view_handler.get_current_view_context()   
        #     if self.view_handler._does_name_represent_web_view(view_name) :
        #         self.view_handler.switch_to_native_view() 
        #         switch_view_context = view_name

        self.dispatcher.take_screenshot(file_path)

        # if switch_view_context:
        #     self.view_handler.switch_to_view_context(switch_view_context)

    def take_screenshot(self, prefix=None):
        image_b64 = self.dispatcher.take_screenshot_as_base64()
        image = base64.b64decode(image_b64)

        ts = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S-%f")[:-3]
        if prefix:
            import re
            prefix = re.sub(r"\[.*?\]", "", prefix)
            prefix = prefix.replace(".py", "").replace("..", "").replace("::",".").replace(":", ".").replace("/", ".").replace("\\", ".") + "-"
        else:
            prefix = ""
        file_name = "{}{}.png".format(prefix, ts)
        fpath = os.path.join(self.screenshots_dir, file_name)
        f = open(fpath, "wb")
        f.write(image)
        f.close()
        img = Image(fpath=fpath, b64=image_b64)

        from arjuna import Arjuna
        Arjuna.get_report_metadata().add_image(img)
        return img

    def focus_on_main_window(self):
        self.main_window.focus()

    @property
    def source(self) -> GuiPageSource:
        raw_source = self.get_source_from_remote()
        source_parser = GuiPageSource(raw_source)
        source_parser._load()
        return source_parser

    def perform_action_chain(self, single_action_chain):
        from arjuna.interact.gui.auto.automator.actions import SingleActionChain
        action_chain = SingleActionChain(self)
        action_chain.perform(single_action_chain)

    def _find_element_with_js(self, js):
        return self.dispatcher.find_element_with_js(js)

    def _find_multielement_with_js(self, js):
        return self.dispatcher.find_multielement_with_js(js)
    '''
        Public API
    '''

    #### Element Finding

    def element(self, gui, wmd):
        from arjuna.tpi.guiauto.widget.element import GuiElement
        gui_element = GuiElement(gui, wmd) 
        self._load_element(gui_element)
        return gui_element

    def multi_element(self, gui, wmd):
        from arjuna.tpi.guiauto.widget.multielement import GuiMultiElement
        m_guielement = GuiMultiElement(gui, wmd)
        self._load_multielement(m_guielement)
        return m_guielement

    def dropdown(self, gui, wmd):
        from arjuna.tpi.guiauto.widget.dropdown import GuiDropDown
        return GuiDropDown(gui, wmd)

    def radio_group(self, gui, wmd):
        from arjuna.tpi.guiauto.widget.radio_group import GuiRadioGroup
        return GuiRadioGroup(gui, wmd)

    def execute_javascript(self, js, *args):
        return self.browser.execute_javascript(js, 
        *[
            isinstance(arg, GuiElement) and arg.dispatcher or arg for arg in args
        ]
        )

    ################################
    # Components
    ################################

    @property
    def alert(self):
        return self.alert_handler.create_alert()


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

from selenium.webdriver.common.by import By
from selenium.common.exceptions import NoSuchElementException
from arjuna.tpi.error import *
from arjuna.core.error import *
from selenium.webdriver.support.relative_locator import RelativeBy
from selenium.webdriver.remote.webelement import WebElement
from arjuna.tpi.tracker import track

@track("debug")
class SeleniumElementFinder:
    BY_MAP = {
        "ID": By.ID,
        "NAME": By.NAME,
        "TAG_NAME": By.TAG_NAME,
        "CLASS_NAME": By.CLASS_NAME,
        "LINK_TEXT": By.LINK_TEXT,
        "PARTIAL_LINK_TEXT": By.PARTIAL_LINK_TEXT,
        "CSS_SELECTOR": By.CSS_SELECTOR,
        "XPATH": By.XPATH
    }

    RELATION_MAP = {
        "above": "above",
        "below": "below",
        "right_of": "to_right_of",
        "left_of": "to_left_of",
        "near": "near",
    }
    
    @classmethod
    def __create_relative_by(cls, byType, byValue, relations):
        from arjuna import log_debug
        relative_by = RelativeBy({byType: byValue})
        for k,v in relations.items():
            try:
                getattr(relative_by, cls.RELATION_MAP[k.lower()])(v)
            except:
                pass
        log_debug("Root" + str(relative_by.root))
        log_debug("Filters" + str(relative_by.filters))
        return relative_by
        
    # @classmethod
    # def find_element(cls, container, byType, byValue, *, relations=None):
    #     from arjuna import log_debug
    #     log_debug(f"Finding element in container:{container} with wtype:{byType} and wvalue:{byValue} with relations: {relations}")
    #     try:
    #         byType = cls.BY_MAP[byType.upper()]
    #         if not relations:
    #             return container.find_element(byType, byValue)
    #         else:
    #             # Currently Selenium supports Relative locator only for find_elements and at driver level
    #             # For nested element finding change to WebDriver instance if relations are defined.
    #             if isinstance(container, WebElement):
    #                 container = container.parent
    #             elements = container.find_elements(cls.__create_relative_by(byType, byValue, relations))
    #             if len(elements) == 0:
    #                 raise Exception("No elements found.")
    #             else:
    #                 return elements[0]
    #     except Exception as e:
    #         raise GuiWidgetNotFoundError("By.{}={}".format(byType, byValue))

    _POS = {
        "first" : 1,
        "last" : -1
    }

    @classmethod
    def find_element(cls, container, byType, byValue, *, relations=None, filters=None):
        from arjuna import log_debug
        log_debug(f"Finding element in container:{container} with wtype:{byType} and wvalue:{byValue} with relations: {relations} and filters: {filters}")
        elements = cls.find_elements(container, byType, byValue, relations=relations, filters=filters)
        return elements[0]

    @classmethod
    def find_elements(cls, container, byType, byValue, *, relations=None, filters=None):
        from arjuna import log_debug
        log_debug(f"Finding elements in container:{container} with wtype:{byType} and wvalue:{byValue} with relations: {relations} and filters: {filters}")
        sbyType = cls.BY_MAP[byType.upper()]
        if not relations:
            elements = container.find_elements(sbyType, byValue)
        else:
            # Currently Selenium supports Relative locator only for find_elements and at driver level
            # For nested element finding change to WebDriver instance if relations are defined.
            if isinstance(container, WebElement):
                container = container.parent
            rby = cls.__create_relative_by(sbyType, byValue, relations)
            log_debug("Selenium find_elements call made with RelativeBy: {}".format(rby.to_dict()))
            elements = container.find_elements(rby)         
        if len(elements) == 0:
            raise GuiWidgetNotFoundError("By.{}={}".format(byType, byValue), message="No element found for Selenium locator")   
        if filters is not None:
            if "pos" in filters:
                log_debug("Filtering elements with filter: {}".format(filters["pos"]))
                try:
                    extracted_elements = filters["pos"].extract(elements)
                    if type(extracted_elements) is not list:
                        return [extracted_elements]
                    else:
                        return extracted_elements
                except Exception as e:
                    log_debug("Exception in filters processing: {}".format(e))
                    raise GuiWidgetNotFoundError("By.{}={}".format(byType, byValue), message="No element found for Selenium locator after applying position filters.")   
        return elements



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.interact.http.model.internal.helper.yaml import convert_yaml_obj_to_content

class HttpRequestYamlRepr:

    def __init__(self, session, req_yaml, *, label):
        self.__label = label
        if "method" in req_yaml:     
            self.__method = req_yaml["method"]
            del req_yaml["method"]
        else:
            self.__method = "get"

        if "route" in req_yaml:     
            self.__route = req_yaml["route"]
            del req_yaml["route"]
        else:
            self.__route = "/"

        if "content_type" in req_yaml:
            from arjuna import Http
            content_handler = getattr(Http.content, req_yaml["content_type"].lower())
            del req_yaml["content_type"]
        else:
            content_handler = session.request_content_handler

        if "content" in req_yaml:
            req_yaml["content"] = convert_yaml_obj_to_content(req_yaml["content"])
            # if req_yaml["content"] is None:
            #     req_yaml["content"] == "null"

            req_yaml["content"] = content_handler(req_yaml["content"])

        self.__attrs = req_yaml

    @property
    def label(self):
        return self.__label

    @property
    def method(self):
        return self.__method

    @property
    def route(self):
        return self.__route

    @property
    def attrs(self):
        return self.__attrs



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite3


class BaseDB:
    def __init__(self, db_path=None):
        self.db_path = db_path
        self.connection = sqlite3.connect(self.db_path)
        self.cursor = self.connection.cursor()

    def drop_table(self, table_name):
        try:
            self.cursor.execute("DROP TABLE %s" % table_name)
        except:
            pass

    def execute(self, sql):
        # print self.cursor
        results = self.cursor.execute(sql)
        return results

    def commit(self):
        self.connection.commit()

    def close(self):
        self.connection.close()



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .gui import Gui
from .content import GuiAppContent
from arjuna.interact.gui.auto.finder.wmd import GuiWidgetMetaData
from arjuna.tpi.tracker import track
from arjuna.tpi.guiauto.source.element import GuiElementSource

@track("debug")
class GuiSection(GuiAppContent):
    '''
        Represents a GUI Section i.e. a part of the current page in the Gui. 
        
        It is an implementation of **GuiAppContent**.

        Args:
            *args: Any number of positional argumnts. These are passed to the **prepare()** method if defined in inherited class.

        Keyword Arguments:
            parent_gui: (Mandatory) The **Gui** object that contains this **GuiSection**.
            label: Label for the this **GuiSection**. If not provided, the class name is used as the label.
            root: Root element of this **GuiSection**. Can be **label** string defined in its GNS File or a **Locator** object.
            gns_dir: Relative Root Directory for GNS file associated with this **GuiSection**. Default is **page/section** directory in associated **GuiApp** namespace. If provided, it is considered relative to the namespace directory of associated **GuiApp**.
            gns_file_name: Name of GNS file associated with this **GuiSection**. If not provided, default is **<label>.yaml**.
            kwargs: Arbitrary keyword arugments. These are passed to the **prepare()** method if defined in inherited class.

        Note:
            A **GuiSection** can have a root element. If defined, all locator calls in this object happen as a nested locating call using the root element.

                - Root element provided in in __init__ call is given preference.
                - If not provided, Arjuna looks for **root** definition in **load** section of its GNS file.
                - If above is also not provided, then locating happens from the root of the DOM of current page.
    '''

    def __init__(self, *args, parent_gui: Gui, label: 'str'=None, root: 'LabelOrLocator'=None, gns_dir: str=None, gns_file_name:str=None, **kwargs):
        super().__init__(automator=parent_gui._automator, label=label, gns_dir=gns_dir, gns_file_name=gns_file_name)   
        self.__root_meta = self.__determine_root(root)
        self.__root_element = None
        self.__container = self
        self._load(*args, **kwargs)
        self.__parent = parent_gui

    @property
    def _root_element(self):
        return self.__root_element

    def __determine_root(self, root_init):
        from arjuna import GuiWidgetDefinition
        root_label = None
        root_gns = self._gui_def.root_element_name
        if root_init:
            # root in __init__ as a Locator instead of GNS Label
            if isinstance(root_init, GuiWidgetDefinition):
                root_label = "anonymous"
            else:
                root_label = root_init
        else:
            root_label = root_gns
        from arjuna import log_debug
        log_debug("Setting Root Element for {} Gui. Label: {}. Root in GNS: {}. Root in __init__: {}.".format(
            self.label,
            root_label,
            root_gns,
            root_init
        ))

        if root_label:
            root_label = root_label.lower().strip()
            return root_label, root_init
        else:
            return None

    def _load_root_element(self):
        '''
            Loads root element for GuiSection.

            Root element is always loaded by using GuiPage. Rest of the elements are loaded as nested elements in root.
        '''
        if self.__root_meta:
            label, locator = self.__root_meta
            if self.__root_meta[0] != "anonymous":
                wmd = self._gui_def.get_wmd(label)
                from arjuna import log_debug
                log_debug("Loading Root Element {} for Gui GuiSection: {}".format(label, self.label))
                self.__root_element = self._wmd_finder.element(wmd)
            else:
                from arjuna import log_debug
                log_debug("Loading Root Element with Locator {} for Gui GuiSection: {}".format(str(locator), self.label))
                self.__root_element = self._finder.locate(locator)           
            
            self.__container = self.__root_element

    def __get_caller(self, name):
        if self.__container is self:
            return getattr(super(), name)
        else:
            return getattr(self.__container, name)

    def locate(self, locator):
        '''
           Locate a GuiWidget.

           Arguments:
            locator: `GuiWidgetDefinition` object.

            Returns:
                An object of type `GuiWidget`. Exact object type depends on the value of **type** attribute in `GuiWidgetDefinition`. 
        '''
        return self.__get_caller("locate")(fargs=fargs, **kwargs)

    def element(self, *, fargs=None, **kwargs) -> 'GuiElement':
        '''
            Locate a `GuiElement`.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetDefinition`. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetDefinition`

            Returns:
                `GuiElement` object.
        '''
        return self.__get_caller("element")(fargs=fargs, **kwargs)

    def multi_element(self, fargs=None, **kwargs) -> 'GuiMultiElement':
        '''
            Locate a `GuiMultiElement`.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetDefinition`. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetDefinition`

            Returns:
                `GuiMultiElement` object.
        '''
        return self.__get_caller("multi_element")(fargs=fargs, **kwargs)

    def dropdown(self, fargs=None, **kwargs) -> 'GuiDropDown':
        '''
            Locate a `GuiDropDown`.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetDefinition`. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetDefinition`

            Returns:
                `GuiDropDown` object.
        '''
        return self.__get_caller("dropdown")(fargs=fargs, **kwargs)

    def radio_group(self, fargs=None, **kwargs) -> 'GuiRadioGroup':
        '''
            Locate a `GuiRadioGroup`

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetDefinition`. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetDefinition`

            Returns:
                `GuiRadioGroup` object
        '''
        return self.__get_caller("radio_group")(fargs=fargs, **kwargs)     

    @property
    def parent(self):
        '''
            Parent GUI of this **GuiSection**.
        '''
        return self.__parent

    @property
    def source(self) -> GuiElementSource:
        '''
           `GuiElementSource` object for the root element of this **GuiSection**.

           Raises:
            Exception is raised if root element is not defined.
        '''
        return self._automator.source


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, auto

class ModifierKey(Enum):
    CTRL = auto()
    CMD = auto()
    XCTRL = auto()
    ALT = auto()
    SHIFT = auto()

class DesktopOS(Enum):
    WINDOWS = auto()
    MAC = auto()
    LINUX = auto()

class SetuActorMode(Enum):
    LOCAL = auto()
    REMOTE = auto()

class MobileOsName(Enum):
    ANDROID = auto()
    IOS = auto()

class AppiumAndroidBrowserName(Enum):
    BROWSER = auto()
    CHROME = auto()

class AppiumIosBrowserName(Enum):
    SAFARI = auto()

class FileFormat(Enum):
    INI = auto()
    TXT = auto()
    DELIMITED = auto()
    XLS = auto()
    CSV = auto()
    GNS = auto()

class DataFileFormat(Enum):
    INI = auto()
    TXT = auto()
    DELIMITED = auto()
    XLS = auto()
    CSV = auto()

class Order(Enum):
    RETAIN = auto()
    BY_NAME = auto()
    RANDOM = auto()

class Filter(Enum):
    INCLUDE = auto()
    EXCLUDE = auto()

class GuiAutomationContext(Enum):
    WEB = auto()
    NATIVE = auto()
    SCREEN = auto()
    ANDROID_WEB = auto()
    IOS_WEB = auto()
    ANDROID_NATIVE = auto()
    IOS_NATIVE = auto()

    # DESKTOP_CONTEXTS = {GuiAutomationContext.NATIVE, GuiAutomationContext.WEB}
    # MOBILE_WEB_CONTEXTS = {GuiAutomationContext.ANDROID_WEB, GuiAutomationContext.IOS_WEB}
    # ALL_WEB_CONTEXTS = {GuiAutomationContext.WEB, GuiAutomationContext.ANDROID_WEB, GuiAutomationContext.IOS_WEB}
    # MOBILE_NATIVE_CONTEXTS = {GuiAutomationContext.ANDROID_NATIVE, GuiAutomationContext.IOS_NATIVE}
    
    @staticmethod
    def isDesktopContext(context):
        return context in GuiAutomationContext.DESKTOP_CONTEXTS

    @staticmethod
    def is_mobile_web_context(context):
        return context in GuiAutomationContext.MOBILE_WEB_CONTEXTS

    @staticmethod
    def is_mobile_native_context(context):
        return context in GuiAutomationContext.MOBILE_NATIVE_CONTEXTS

    @staticmethod
    def is_web_context(context):
        return context in GuiAutomationContext.ALL_WEB_CONTEXTS

class GuiAutomatorName(Enum):
    SELENIUM = auto()
    APPIUM = auto()

class GuiElementType(Enum):
    TEXTBOX = auto()
    PASSWORD = auto()
    LINK = auto()
    BUTTON = auto()
    SUBMIT_BUTTON = auto()
    DROPDOWN = auto()
    CHECKBOX = auto()
    RADIO = auto()
    IMAGE = auto()

class OS(Enum):
	WINDOWS = auto()
	MAC = auto()
	LINUX = auto()
	ANDROID = auto()
	IOS = auto()

class MobileView(Enum):
    NATIVE_APP = auto()
    WEBVIEW = auto()

class Device(Enum):
    PC = auto()
    MOBILE = auto()
    GENERIC = auto()    

class Device(Enum):
    UP = auto()
    DOWN = auto()
    LEFT = auto()
    RIGHT = auto()


class ConfigPropertyFormattingTypeEnum(Enum):
    PATH_TO_ABS_PATH = auto()

class DeviceTypeEnum(Enum):
    PC = auto()
    MOBILE = auto()
    GENERIC = auto()

class HoconSyntaxTypeEnum(Enum):
    PROPERTIES = auto()
    JSON = auto()
    CONF = auto()


class FilterTypeEnum(Enum):
    INCLUDE = auto()
    EXCLUDE = auto()


class DiscoveredFileAttributeEnum(Enum):
    NAME = auto()
    EXTENSION = auto()
    FULL_NAME = auto()
    DIRECTORY_ABSOLUTE_PATH = auto()
    DIRECTORY_RELATIVE_PATH = auto()
    PACKAGE_DOT_NOTATION = auto()
    COMMA_SEPATARED_RELATIVE_PATH = auto()
    CONTAINER = auto()
    CONTAINER_TYPE = auto()

class ValueType(Enum):
    BOOLEAN = auto()
    STRING = auto()
    STRING_LIST = auto()
    NONE = auto()
    NUMBER = auto()
    NUMBER_LIST = auto()
    LIST = auto()
    ANYREF = auto()
    ENUM = auto()
    ENUM_LIST = auto()
    INTEGER = auto()
    OBJECT_LIST = auto()
    FLOAT = auto()
    DOUBLE = auto()
    LONG = auto()
    NOT_SET = auto()
    NA = auto()
    INT_LIST = auto()

class NamesContainerTypeEnum(Enum):
    TEST = auto()
    TEST_RESULT = auto()
    IGNORED_TEST = auto()
    STEP_RESULT = auto()
    ISSUE = auto()
    DEFAULT_FIXTURE_tfuncs = auto()
    COMPONENT_NAMES = auto()
    TEST_OBJECT = auto()
    EXCLUDED_TEST_RESULT = auto()
    EVENT = auto()
    TEST_OTYPE_NAMES = auto()
    FIXTURE_RESULT = auto()

class ConfigPropertyLevelEnum(Enum):
    CENTRAL = auto()
    THREAD = auto()

class CorePropertyTypeEnum(Enum):
    ARJUNA_ROOT_DIR = auto()
    PROG = auto()
    CONFIG_CENTRAL_FILE_NAME = auto()
    CONFIG_PROJECTS_DIR = auto()
    WORKSPACE_DIR = auto()
    EXTERNAL_TOOLS_DIR = auto()
    EXTERNAL_IMP_DIR = auto()
    LOGGER_DIR = auto()
    CONFIG_DIR = auto()
    LOGGER_CONSOLE_LEVEL = auto()
    LOGGER_FILE_LEVEL = auto()
    LOGGER_NAME = auto()
    PROJECT_DIRS_FILES = auto()

class GuiInteractionConfigType(Enum):
    CHECK_TYPE = auto()
    CHECK_PRE_STATE = auto()
    CHECK_POST_STATE = auto()
    SCROLL_TO_VIEW = auto()

class ReportFormat(Enum):
    XML = auto()
    HTML = auto()

class DataRefType(Enum):
    CONTEXTUAL = auto()
    INDEXED = auto()

class LoggingLevel(Enum):
    TRACE = auto()
    DEBUG = auto()
    INFO = auto()
    WARN = auto()
    ERROR = auto()
    FATAL = auto()

class GuiWidgetType(Enum):
    ELEMENT = auto()
    MULTI_ELEMENT = auto()
    DROPDOWN = auto()
    RADIO_GROUP = auto()

class DryRunType(Enum):
    SHOW_TESTS = auto()
    SHOW_PLAN = auto()
    CREATE_RES = auto()

class BuiltInProp(Enum):
    PACKAGE = auto()
    MODULE = auto()
    QUAL_NAME = auto()
    ID = auto()
    PRIORITY = auto()
    THREADS= auto()
    NAME = auto()
    AUTHOR = auto()
    IDEA = auto()
    UNSTABLE = auto()
    COMPONENT = auto()
    APP_VERSION = auto()
    LEVEL = auto()
    REVIEWED = auto()

class RuleNature(Enum):
    INCLUDE = auto()
    EXCLUDE = auto()

class RuleType(Enum):
    SET = auto()
    DICT_KEY = auto()
    DICT_VALUE = auto()

class RuleTargetType(Enum):
    PROPS = auto()
    TAGS = auto()
    BUGS = auto()

class RuleConditionType(Enum):
    EQUAL = auto()
    NOT_EQUAL = auto()
    MATCHES = auto()
    DOES_NOT_MATCH = auto()
    PARTIALLY_MATCHES = auto()
    DOES_NOT_PARTIALLY_MATCH = auto()
    LESS_THAN = auto()
    LESS_OR_EQUAL = auto()
    GREATER_THAN = auto()
    GREATER_OR_EQUAL = auto()
    CONTAINS = auto()
    IS_SUBSET = auto()
    HAS_INTERSECTION = auto()
    NO_INTERSECTION = auto()

class ConfigStage(Enum):
    DEFAULT = auto()
    REFERENCE = auto()
    PROJECT = auto()
    CODED = auto()
    CLI = auto()


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

##########################################################
# Raised and consumed by internal implementation of Arjuna
##########################################################

import re

def arj_format_str(target, vargs, repl_dict):

    def get_global_value(in_str):
        from arjuna import C, L, R
        gtype, query = in_str.split(".", 1)
        gtype = gtype.upper()
        return locals()[gtype](query)

    do_eval = False
    if type(target) is not str:
        do_eval = True
        target = str(target)

    pos_pattern = r"(\$\s*\$)"
    named_pattern = r"\$(\s*[\w\.]+?\s*)\$"
    fmt_target = target.replace("{", "__LB__").replace("}", "__RB__")

    # Find params
    pos_matches = re.findall(pos_pattern, fmt_target)
    named_matches = re.findall(named_pattern, fmt_target)

    if pos_matches and named_matches:
        for match in named_matches:
            match = match.lower().strip()
            if not match.startswith("c.") and not match.startswith("l.") and not match.startswith("r."):
                raise Exception("You can not use positional $$ placeholders and named $<name>$ placholders together in a withx locator definition (except those with C./L./R. prefixes to fetch global values. Wrong withx definition: " + target)

    if pos_matches:
        if not vargs:
            vargs = tuple()    
        if len(pos_matches) != len(vargs):
            raise Exception("Number of positional arguments supplied to format withx locator do not match number of $$ placeholders. Placeholders: {}. Positional args: {}. Wrong usage of withx locator: {}".format(len(pos_matches), vargs, target))

        for i,match in enumerate(pos_matches):
            fmt_target = fmt_target.replace(match, "{}")

        fmt_target = fmt_target.format(*vargs)
        
    for match in named_matches:
        names_set = None
        target = "${}$".format(match)
        processed_name = match.lower().strip()
        repl_value = None
        if processed_name.startswith("c.") or processed_name.startswith("l.") or processed_name.startswith("r."):
            repl_value = get_global_value(processed_name)
        else:
            if processed_name not in repl_dict:
                continue
            repl_value = repl_dict[processed_name]

        fmt_target = fmt_target.replace(target, str(repl_value))

    fmt_target = fmt_target.replace("__LB__", "{").replace("__RB__", "}")
    if do_eval:
        fmt_target = eval(fmt_target)
    return fmt_target

def arj_convert(content):

        if content is None or content == "":
            return content

        from json import JSONEncoder, dumps, loads
        from arjuna.tpi.parser.yaml import YamlDict, YamlList
        from arjuna.tpi.parser.json import JsonDict, JsonList
        from arjuna.tpi.data.entity import _DataEntity

        class _CustomEncoder(JSONEncoder):
            def default(self, o):
                if isinstance(o, YamlDict) or isinstance(o, YamlList) or isinstance(o, JsonList) or isinstance(o, JsonDict):
                    return o.raw_object
                elif isinstance(o, _DataEntity):
                    return o.as_dict()
                return JSONEncoder.default(self, o)

        if type(content) is tuple:
            content = list(content)

        if content:
            if isinstance(content, JsonList) or isinstance(content, JsonDict):
                content = content.raw_object
            elif isinstance(content, _DataEntity):
                content = content.as_dict()
            elif content == "null":
                content = None
            content = dumps(content, cls=_CustomEncoder, indent=2)
        else:
            if content is not None:
                content = str(content)

        if content != "":
            return loads(content)
        else:
            return ""


import re
from enum import Enum, auto
from .ref import *

from arjuna.core.constant import *
from arjuna.core.error import *

class Pattern:

    @classmethod
    def extract(cls, pattern, target_str):
        return pattern.match(target_str)


class Rule:

    def __init__(self, rule_str, container, target, condition, expression):
        self.__rule_str = rule_str
        self.__container = container
        self.__target = target
        self.__condition = condition
        self.__expression = expression  
        self.__expression_type = type(expression)
        self.__checker = get_value_checker_for_symbol(self.__condition)

    @property
    def rule_str(self):
        return self.__rule_str

    @property
    def container(self):
        return self.__container

    @property
    def target(self):
        return self.__target

    @property
    def condition(self):
        return self.__condition

    @property
    def expression(self):
        return self.__expression    

    @property
    def expression_type(self):
        return self.__expression_type 

    @property
    def checker(self):
        return self.__checker 

    def _get_container_obj(self, obj):
        return getattr(obj, self.container)

    def __str__(self):
        return "{}(rule_str='{}', container={}, target={}, condition={}, expression={}, expression_type={}, checker={})".format(self.__class__.__name__, self.rule_str, self.container, self.target, self.condition, self.expression, self.expression_type.__name__, self.checker.__name__)

class BoolAttrPatternRule(Rule):
    '''
        Simple Pattern for Boolean test attributes.

        For example:

            unstable
            not unstable
    '''

    __NOT_PATTERN = re.compile(r"^\s*not\s+(?P<target>\w+)\s*$",re.IGNORECASE)
    __PATTERN = re.compile(r"^\s*(?P<target>\w+)\s*$",re.IGNORECASE)

    def __init__(self, *, rule_str, target, condition, expression):
        super().__init__(rule_str, "info", target, condition, expression)

    @classmethod
    def _validate(cls, target):
        if is_builtin_prop(target) and get_value_type(target) is not bool:
            raise InvalidSelectionRule("Built-in property {} is not of bool type. It can not be used in a {} construct.".format(target, cls.__name__))

    @classmethod
    def from_str(cls, rule_str):  
        match = Pattern.extract(cls.__NOT_PATTERN, rule_str)
        if match:
            target = match.group('target')
            cls._validate(target)
            return BoolAttrPatternRule(**{
                'rule_str':rule_str, 
                'target': target,
                'condition' : RuleConditionType.NOT_EQUAL,
                'expression' : True
            })
        else:
            match = Pattern.extract(cls.__PATTERN, rule_str)
            if match:
                target = match.group('target')
                cls._validate(target)
                return BoolAttrPatternRule(**{
                'rule_str':rule_str, 
                'target': target,
                'condition' : RuleConditionType.EQUAL,
                'expression' : True
                })
            else:
                raise RulePatternDoesNotMatchError(rule_str, cls, "[not] <boolean_object>")

    def matches(self, obj):
        container = self._get_container_obj(obj)
        if not hasattr(container, self.target):
            actual = False
        else:
            actual = getattr(container, self.target)

        if is_builtin_prop(self.target):
            return self.checker(actual, self.expression)
        else:
            # For user define attrs, the value in @test is converted to bool
            return self.checker(bool(actual), self.expression)


class IterablePatternRule(Rule):
    '''
        Simple Pattern for presence/absence of strings in iterables.

        For example:

            with tags a,b
            without tags a,b
            with bugs a,b
            without bugs a,b
            with envs a,b
            without envs a,b

            with usertags a,b (found in info rather than object)

            You can use singular version as well - **tag/bug/env**
    '''

    __WITH_PATTERN = re.compile(r"^\s*with\s+(?P<container>(.+?))\s+(?P<tags>(.*))$",re.IGNORECASE)
    __WITHALL_PATTERN = re.compile(r"^\s*withall\s+(?P<container>(.+?))\s+(?P<tags>(.*))$",re.IGNORECASE)
    __WITHOUT_PATTERN = re.compile(r"^\s*without\s+(?P<container>(.+?))\s+(?P<tags>(.*))$",re.IGNORECASE)

    def __init__(self, *, rule_str, container, target, condition, expression):
        super().__init__(rule_str, container, target, condition, expression)

    @property
    def tags(self):
        return self.__tags

    @classmethod
    def __convert_to_tags(cls, expression):
        return set([p.strip().lower() for p in expression.split(",")])

    @classmethod
    def from_str(cls, rule_str):  
        condition = None
        match = Pattern.extract(cls.__WITH_PATTERN, rule_str)
        if match:
            condition = RuleConditionType.HAS_INTERSECTION
        else:
            match = Pattern.extract(cls.__WITHALL_PATTERN, rule_str)
            if match:
                condition = RuleConditionType.IS_SUBSET
            else:
                match = Pattern.extract(cls.__WITHOUT_PATTERN, rule_str)
                if match:
                    condition = RuleConditionType.NO_INTERSECTION
                else:
                    raise RulePatternDoesNotMatchError(rule_str, cls, "[with|withall|without]  <tag-container> tag1[,....]")

        tags = cls.__convert_to_tags(match.group('tags'))
        container = get_tag_container(match.group('container'))
        return IterablePatternRule(**{
            'rule_str':rule_str,
            'container' : container,
            'target': tags, 
            'condition' : condition,
            'expression': tags
        })        

    def matches(self, obj):
        try:
            container = self._get_container_obj(obj)
        except AttributeError:
            # This logic block is executed when the rule looks for an iterable with name self.container in obj.info
            try:
                info_obj = getattr(obj, 'info')
                container = getattr(info_obj, self.container)
                if container is None:
                    container = set()
                elif type(container) == str:
                    container = {container}
                else:
                    try:
                        container = set([str(i) for i in container])
                    except:
                        container = {str(container)}
            except:
                # For user defined tag container.
                container = set()

        if container is None:
            container = set()
        if not container and self.condition == RuleConditionType.IS_SUBSET:
            return False
        return self.checker(self.expression, container)


class AttrPatternRule(Rule):
    '''
        Pattern for executing a condition on an attribute value.

        For example:

            author is Rahul
    '''
    __condition_str = "|".join(get_all_symbols()).replace('*=',r'\*=') #"is|not|eq|=|==|!=|ne|lt|<|le|<=|gt|>|ge|>=|matches|~=|contains|\*="
    __p5_raw = r"^\s*(?P<target>\w+)\s+(?P<condition>({}){{1,1}})\s+(?P<expression>.*?)\s*$".format(__condition_str)

    __PATTERN = re.compile(__p5_raw, re.IGNORECASE)

    def __init__(self, *, rule_str, target, condition, expression):
        def remove_py_ext(name):
            if not name.lower().endswith(".py"):
                return name
            else:
                return name.replace(".py","")
        if target == "module":
            expression = remove_py_ext(expression)
        super().__init__(rule_str, "info", target, condition, expression)

    @classmethod
    def from_str(cls, rule_str):  
        match = Pattern.extract(cls.__PATTERN, rule_str)
        if match:
            target = match.group('target')
            # For rule matching in info-pattern, target is always a str type for user-defined attrs.
            if is_builtin_prop(target):
                target_type = get_value_type(target) 
            else:
                target_type = str
            return AttrPatternRule  (**{
                'rule_str':rule_str,                 
                'target': target, 
                'condition' : convert_to_condition(target, target_type, match.group('condition')), 
                'expression': convert_expression(target, target_type, match.group('expression')),
            })
        else:
            raise RulePatternDoesNotMatchError(rule_str, cls, "<prop_name> <condition> <expression>")

    def matches(self, obj):
        container = self._get_container_obj(obj)
        if not hasattr(container, self.target):
            actual = get_default_value_for_type(self.expression_type, self.target)
        else:
            actual = getattr(container, self.target)

        if is_builtin_prop(self.target):
            return self.checker(actual, self.expression)
        else:
            # User attr in property rule is always str-to-str
            # If actual value in @test is None for a user property, the rule is considered not matched.
            if actual is None:
                return False
            return self.checker(str(actual), self.expression)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.helper.arjtype import CIStringDict

from arjuna.tpi.constant import *
from arjuna.core.constant import *
from arjuna.core.utils.repr_utils import repr_dict
from arjuna.tpi.tracker import track

class InteractionConfig:

    def __init__(self, meta_dict):
        self.__settings = {
            GuiInteractionConfigType.CHECK_TYPE: True,
            GuiInteractionConfigType.CHECK_PRE_STATE : True,
            GuiInteractionConfigType.CHECK_POST_STATE : True,
            GuiInteractionConfigType.SCROLL_TO_VIEW : False,
        }   

        loaded = []
        for k,v in meta_dict.items():
            try:
                self[k] = v
                loaded.append(k)
            except:
                pass

        for l in loaded:
            del meta_dict[l]

    def should_check_type(self):
        return self.__settings[GuiInteractionConfigType.CHECK_TYPE]

    def should_check_pre_state(self):
        return self.__settings[GuiInteractionConfigType.CHECK_PRE_STATE]

    def should_check_post_state(self):
        return self.__settings[GuiInteractionConfigType.CHECK_POST_STATE]

    def should_scroll_to_view(self):
        return self.__settings[GuiInteractionConfigType.SCROLL_TO_VIEW]

    def __setitem__(self, name, value):
        if isinstance(name, GuiInteractionConfigType):
            self.__settings[k] = value
        else:
            self.__settings[GuiInteractionConfigType[name.upper()]] = value

    @classmethod
    def is_a_setting(cls, name):
        if isinstance(name, GuiInteractionConfigType):
            return True
        else:
            try:
                GuiInteractionConfigType[name.upper()]
                return True
            except:
                return False


    def __str__(self):
        return repr_dict(self.__settings)

class Meta:
    _ALLOWED_FILTERS = {
        GuiWidgetType.ELEMENT: {'pos'},
        GuiWidgetType.MULTI_ELEMENT: {'pos'},
        GuiWidgetType.DROPDOWN: {'pos'},
        GuiWidgetType.RADIO_GROUP: set(),
    }

    _POS = {"first", "last", "random", "odd", "even"}

    @track("trace")
    def __init__(self, mdict=None):
        from arjuna import log_debug
        log_debug("Input Meta Dict for Meta creation: {}".format(repr_dict(mdict)))
        temp_dict = not mdict and CIStringDict() or CIStringDict(mdict)
        self.__mdict = CIStringDict()
        self.__process_type(temp_dict)
        self.__process_relations(temp_dict)
        self.__process_filters(temp_dict)
        self.__process_settings(temp_dict)
        self.__mdict.update({k:v for k,v in temp_dict.items() if k.lower() not in {"type", "relations", "settings", "filters"}})
        print(self.__mdict)
        log_debug("Meta dictionary is: {}".format(repr_dict(self.__mdict)))

    def __process_type(self, temp_dict):
        from arjuna import log_debug
        from arjuna.core.constant import GuiWidgetType
        if "type" in temp_dict:
            log_debug("Copying provided type from meta dict: {}".format(temp_dict["type"]))
            try:
                widget_type = temp_dict["type"]
                if not isinstance(widget_type, GuiWidgetType):
                    self.__mdict["type"] = GuiWidgetType[widget_type.upper()]
                else:
                    self.__mdict["type"] = temp_dict["type"]
            except Exception as e:
                raise Exception("{} is not a valid Gui widget type.".format(widget_type))
        else:
            self.__mdict["type"] = GuiWidgetType.ELEMENT

    def __process_relations(self, temp_dict):
        from arjuna import log_debug
        self.__mdict["relations"] = CIStringDict()
        to_remove = list()
        for k,v in temp_dict.items():
            if k.lower() in {'above', 'below', 'left_of', 'right_of', 'near'}:
                 
                to_remove.append(k)
        for k in to_remove:
            del temp_dict[k]

        if "relations" in temp_dict:
            self.__mdict["relations"].update(temp_dict["relations"])
            del temp_dict["relations"]

        from arjuna.tpi.guiauto.base.single_widget import SingleGuiWidget
        from arjuna.tpi.error import GuiWidgetDefinitionError

        for k,v in self.__mdict["relations"].items():
            self.__set_relation(k, v)

    def __process_filters(self, temp_dict):
        from arjuna import log_debug
        self.__mdict["filters"] = CIStringDict()
        to_remove = list()
        for k,v in temp_dict.items():
            if k.lower() in {'pos'}:
                self.__set_filter(k,v)
                to_remove.append(k)
        for k in to_remove:
            del temp_dict[k]
        if "filters" in temp_dict:
            self.__mdict["filters"].update(temp_dict["filters"])
            del temp_dict["filters"]

    def __process_settings(self, temp_dict):
        self.__mdict["settings"] = InteractionConfig(temp_dict) # Interconfig keys are removed

    def update_settings(self, source_wmd):
        self.settings.update(source_wmd.meta.settings)

    def items(self):
        return self.__mdict.items()

    def has(self, name):
        return name.lower() in self.__mdict

    @track("trace")
    def __getattr__(self, name):
        return self[name]

    def __getitem__(self, name):
        from arjuna import log_debug
        if self.has(name):
            return self.__mdict[name]
        else:
            log_debug(f"Meta dict does not have {name}. Returning None.")
            return None

    def __set_relation(self, name, value):
        from arjuna.tpi.guiauto.base.single_widget import SingleGuiWidget
        if isinstance(value, SingleGuiWidget):
            value = value.dispatcher.driver_element
        self.__mdict["relations"][name] = value

    def __format_pos(self, pos):
        from arjuna.tpi.helper.extract import pos as pos_factory
        from arjuna.tpi.helper.extract import Extractor
        if self.__mdict["type"] == GuiWidgetType.RADIO_GROUP:
            raise Exception(">>pos<< filter is not supported for Gui Widget Type RADIO_GROUP.")
        if type(pos) is str:
            fpos = pos.lower().strip()
            if fpos in self._POS:
                return pos_factory._create_extractor(fpos)
            else:
                raise Exception("The only string liternals support for defining position are first/last/random")
        elif type(pos) in {list, tuple}:
            return pos_factory.at(*[int(str(i).strip()) for i in pos])
        elif type(pos) is dict:
            if len(pos) > 1:
                raise Exception("Extractor specification dictionary can take only one root key. Found entry: {}".format(pos))
            extractor_name = list(pos.keys())[0].lower().strip()
            extractor_args = list(pos.values())[0]
            if type(extractor_args) in {list, tuple}:
                return pos_factory._create_extractor(extractor_name, *extractor_args)
            elif type(extractor_args) is dict:
                return pos_factory._create_extractor(extractor_name, **extractor_args)
            else:
                return pos_factory._create_extractor(extractor_name, extractor_args)
        elif isinstance(pos, Extractor):
            return pos
        else:
            try:
                return pos_factory.at(int(str(pos).lower().strip()))
            except:
                raise Exception("Value of pos is not of any allowed type. It can be an int, a list of ints, an extractor specification dictionary or an Extractor object.")          

    def __set_filter(self, name, value):
        allowed_filters = self._ALLOWED_FILTERS[self.__mdict['type']]
        if name not in allowed_filters:
            raise Exception("{} is not allowed filter meta data for GuiWidget of type: {}. Allowed: {}".format(k, self.__mdict['type'], allowed_filters))
        if name == "pos":
            self.__mdict["filters"][name] = self.__format_pos(value)
        else:
            self.__mdict["filters"][name] = value

    def __setitem__(self, name, value):
        if InteractionConfig.is_a_setting(name):
            self["settings"][name] = value
        elif name.lower() in {"above", "below", "near", "right_of", "left_of"}:
            self.__set_relation(name, value)
        elif name.lower() in {"pos", "slice"}:
            self.__set_filter(name, value)
        else:
            self.__mdict[name] = value

    def __str__(self):
        return repr_dict(self.__mdict)


    